"""
Module for optical constants of dust grains.
"""

import os
import zipfile
import urllib.request

import h5py
import numpy as np

from astropy.io import fits
from typeguard import typechecked


@typechecked
def add_optical_constants(input_path: str,
                          database: h5py._hl.files.File) -> None:
    """
    Function for adding the optical constants of crystalline and amorphous MgSiO3 and Fe to the
    database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.

    Returns
    -------
    None
        NoneType
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    url = 'https://home.strw.leidenuniv.nl/~stolker/species/optical_constants.zip'

    data_file = os.path.join(input_path, 'optical_constants.zip')

    if not os.path.isfile(data_file):
        print('Downloading optical constants (87 kB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Unpacking optical constants...', end='', flush=True)

    with zipfile.ZipFile(data_file, 'r') as zip_ref:
        zip_ref.extractall(input_path)

    print(' [DONE]')

    print('Adding optical constants of MgSiO3...', end='')

    nk_file = os.path.join(input_path, 'optical_constants/mgsio3/crystalline/'
                           'mgsio3_jaeger_98_scott_96_axis1.dat')

    data = np.loadtxt(nk_file)
    database.create_dataset('dust/mgsio3/crystalline/axis_1/', data=data)

    nk_file = os.path.join(input_path, 'optical_constants/mgsio3/crystalline/'
                           'mgsio3_jaeger_98_scott_96_axis2.dat')

    data = np.loadtxt(nk_file)
    database.create_dataset('dust/mgsio3/crystalline/axis_2/', data=data)

    nk_file = os.path.join(input_path, 'optical_constants/mgsio3/crystalline/'
                           'mgsio3_jaeger_98_scott_96_axis3.dat')

    data = np.loadtxt(nk_file)
    database.create_dataset('dust/mgsio3/crystalline/axis_3/', data=data)

    nk_file = os.path.join(input_path, 'optical_constants/mgsio3/amorphous/'
                           'mgsio3_jaeger_2003_reformat.dat')

    data = np.loadtxt(nk_file)
    database.create_dataset('dust/mgsio3/amorphous', data=data)

    print(' [DONE]')

    print('Adding optical constants of Fe...', end='')

    nk_file = os.path.join(input_path, 'optical_constants/fe/crystalline/fe_henning_1996.dat')
    data = np.loadtxt(nk_file)
    database.create_dataset('dust/fe/crystalline', data=data)

    nk_file = os.path.join(input_path, 'optical_constants/fe/amorphous/fe_pollack_1994.dat')
    data = np.loadtxt(nk_file)
    database.create_dataset('dust/fe/amorphous', data=data)

    print(' [DONE]')


@typechecked
def add_cross_sections(input_path: str,
                       database: h5py._hl.files.File) -> None:
    """
    Function for adding the extinction cross section of crystalline MgSiO3 for a log-normal and
    power-law size distribution to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.

    Returns
    -------
    None
        NoneType
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    url = 'https://home.strw.leidenuniv.nl/~stolker/species/lognorm_mgsio3_c_ext.fits'

    data_file = os.path.join(input_path, 'lognorm_mgsio3_c_ext.fits')

    print('Downloading log-normal dust cross sections (231 kB)...', end='', flush=True)
    urllib.request.urlretrieve(url, data_file)
    print(' [DONE]')

    print('Adding log-normal dust cross sections:')

    with fits.open(os.path.join(input_path, 'lognorm_mgsio3_c_ext.fits')) as hdu_list:
        database.create_dataset('dust/lognorm/mgsio3/crystalline/cross_section/',
                                data=hdu_list[0].data)

        print(f'   - Data shape (n_wavelength, n_radius, n_sigma): {hdu_list[0].data.shape}')

        database.create_dataset('dust/lognorm/mgsio3/crystalline/wavelength/',
                                data=hdu_list[1].data)

        data_range = f'{np.amin(hdu_list[1].data)} - {np.amax(hdu_list[1].data)}'
        print(f'   - Wavelength range: {data_range} um')

        database.create_dataset('dust/lognorm/mgsio3/crystalline/radius_g/',
                                data=hdu_list[2].data)

        data_range = f'{np.amin(hdu_list[2].data)} - {np.amax(hdu_list[2].data)}'
        print(f'   - Mean geometric radius range: {data_range} um')

        database.create_dataset('dust/lognorm/mgsio3/crystalline/sigma_g/',
                                data=hdu_list[3].data)

        data_range = f'{np.amin(hdu_list[3].data)} - {np.amax(hdu_list[3].data)}'
        print(f'   - Geometric standard deviation range: {data_range}')

    url = 'https://home.strw.leidenuniv.nl/~stolker/species/powerlaw_mgsio3_c_ext.fits'

    data_file = os.path.join(input_path, 'powerlaw_mgsio3_c_ext.fits')

    print('Downloading power-law dust cross sections (231 kB)...', end='', flush=True)
    urllib.request.urlretrieve(url, data_file)
    print(' [DONE]')

    print('Adding power-law dust cross sections')

    with fits.open(os.path.join(input_path, 'powerlaw_mgsio3_c_ext.fits')) as hdu_list:
        database.create_dataset('dust/powerlaw/mgsio3/crystalline/cross_section/',
                                data=hdu_list[0].data)

        print(f'   - Data shape (n_wavelength, n_radius, n_exponent): {hdu_list[0].data.shape}')

        database.create_dataset('dust/powerlaw/mgsio3/crystalline/wavelength/',
                                data=hdu_list[1].data)

        data_range = f'{np.amin(hdu_list[1].data)} - {np.amax(hdu_list[1].data)}'
        print(f'   - Wavelength range: {data_range} um')

        database.create_dataset('dust/powerlaw/mgsio3/crystalline/radius_max/',
                                data=hdu_list[2].data)

        data_range = f'{np.amin(hdu_list[2].data)} - {np.amax(hdu_list[2].data)}'
        print(f'   - Maximum grain radius range: {data_range} um')

        database.create_dataset('dust/powerlaw/mgsio3/crystalline/exponent/',
                                data=hdu_list[3].data)

        data_range = f'{np.amin(hdu_list[3].data)} - {np.amax(hdu_list[3].data)}'
        print(f'   - Power-law exponent range: {data_range}')



"""
Utility functions for reading data.
"""

import math
import warnings

from typing import Dict, Optional, Tuple, Union

import numpy as np

from scipy.integrate import simps
from scipy.ndimage.filters import gaussian_filter
from typeguard import typechecked

from species.core import box, constants
from species.read import read_model, read_planck


@typechecked
def get_mass(logg: Union[float, np.ndarray],
             radius: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    """
    Parameters
    ----------
    logg : float, np.ndarray
        Log10 of the surface gravity (cgs).
    radius : float, np.ndarray
        Radius (Rjup).

    Returns
    -------
    float, np.ndarray
        Mass (Mjup).
    """

    surface_grav = 1e-2 * 10.**logg  # (m s-2)

    radius *= constants.R_JUP  # (m)

    mass = surface_grav*radius**2/constants.GRAVITY  # (kg)
    mass /= constants.M_JUP  # (Mjup)

    return mass


def add_luminosity(modelbox):
    """
    Function to add the luminosity of a model spectrum to the parameter dictionary of the box.

    Parameters
    ----------
    modelbox : species.core.box.ModelBox
        Box with the model spectrum. Should also contain the dictionary with the model parameters,
        the radius in particular.

    Returns
    -------
    species.core.box.ModelBox
        The input box with the luminosity added in the parameter dictionary.
    """

    print('Calculating the luminosity...', end='', flush=True)

    if modelbox.model == 'planck':
        readmodel = read_planck.ReadPlanck(wavel_range=(1e-1, 1e3))
        fullspec = readmodel.get_spectrum(model_param=modelbox.parameters, spec_res=1000.)

    else:
        readmodel = read_model.ReadModel(modelbox.model)
        fullspec = readmodel.get_model(modelbox.parameters)

    flux = simps(fullspec.flux, fullspec.wavelength)

    if 'distance' in modelbox.parameters:
        luminosity = 4.*np.pi*(fullspec.parameters['distance']*constants.PARSEC)**2*flux  # (W)

        # Analytical solution for a single-component Planck function
        # luminosity = 4.*np.pi*(modelbox.parameters['radius']*constants.R_JUP)**2* \
        #     constants.SIGMA_SB*modelbox.parameters['teff']**4.

    else:
        luminosity = 4.*np.pi*(fullspec.parameters['radius']*constants.R_JUP)**2*flux  # (W)

    modelbox.parameters['luminosity'] = luminosity/constants.L_SUN  # (Lsun)

    print(' [DONE]')

    print(f'Wavelength range (um): {fullspec.wavelength[0]:.2e} - '
          f'{fullspec.wavelength[-1]:.2e}')

    print(f'Luminosity (Lsun): {luminosity/constants.L_SUN:.2e}')

    return modelbox


@typechecked
def update_spectra(objectbox: box.ObjectBox,
                   model_param: Dict[str, float],
                   model: Optional[str] = None) -> box.ObjectBox:
    """
    Function for applying a flux scaling and/or error inflation to the spectra of an
    :class:`~species.core.box.ObjectBox`.

    Parameters
    ----------
    objectbox : species.core.box.ObjectBox
        Box with the object's data, including the spectra.
    model_param : dict
        Dictionary with the model parameters. Should contain the value(s) of the flux scaling
        and/or the error inflation.
    model : str, None
        Name of the atmospheric model. Only required for inflating the errors. Otherwise, the
        argument can be set to ``None``.

    Returns
    -------
    species.core.box.ObjectBox
        The input box which includes the spectra with the scaled fluxes and/or inflated errors.
    """

    if objectbox.flux is not None:

        for key, value in objectbox.flux.items():
            if f'{key}_error' in model_param:
                var_add = model_param[f'{key}_error']**2 * value[0]**2

                message = f'Inflating the error of {key} (W m-2 um-1): {np.sqrt(var_add):.2e}...'
                print(message, end='', flush=True)

                value[1] = np.sqrt(value[1]**2 + var_add)

                print(' [DONE]')

            objectbox.flux[key] = value

    if objectbox.spectrum is not None:
        # Check if there are any spectra

        for key, value in objectbox.spectrum.items():
            # Get the spectrum (3 columns)
            spec_tmp = value[0]

            if f'scaling_{key}' in model_param:
                # Scale the flux of the spectrum
                scaling = model_param[f'scaling_{key}']

                print(f'Scaling the flux of {key}: {scaling:.2f}...', end='', flush=True)
                spec_tmp[:, 1] *= model_param[f'scaling_{key}']
                print(' [DONE]')

            if f'error_{key}' in model_param:
                # Calculate the model spectrum
                wavel_range = (0.9*spec_tmp[0, 0], 1.1*spec_tmp[-1, 0])
                readmodel = read_model.ReadModel('atmo', wavel_range=wavel_range)

                model = readmodel.get_model(model_param,
                                            spec_res=value[3],
                                            wavel_resample=spec_tmp[:, 0],
                                            smooth=True)

                # Scale the errors relative to the model spectrum
                err_scaling = model_param[f'error_{key}']
                log_msg = f'Inflating the error of {key}: {err_scaling:.2e}...'

                print(log_msg, end='', flush=True)
                spec_tmp[:, 2] = np.sqrt(spec_tmp[:, 2]**2 + (err_scaling*model.flux)**2)
                print(' [DONE]')

            # Store the spectra with the scaled fluxes and/or errors
            # The other three elements (i.e. the covariance matrix, the inverted covariance matrix,
            # and the spectral resolution) remain unaffected
            objectbox.spectrum[key] = (spec_tmp, value[1], value[2], value[3])

    return objectbox


@typechecked
def create_wavelengths(wavel_range: Tuple[Union[float, np.float32], Union[float, np.float32]],
                       spec_res: float) -> np.ndarray:
    """
    Function for creating logarithmically-spaced wavelengths at a constant spectral resolution.

    Parameters
    ----------
    wavel_range : tuple(float, float)
        Wavelength range (um). Tuple with the minimum and maximum wavelength.
    spec_res : float
        Spectral resolution at which the wavelengths are sampled.

    Returns
    -------
    np.ndarray
        Array with the wavelength points and a fixed spectral resolution. Since the wavelength
        boundaries are fixed, the output spectral resolution is slightly different from the
        ``spec_res`` value.
    """

    n_test = 100

    wavel_test = np.logspace(np.log10(wavel_range[0]), np.log10(wavel_range[1]), n_test)

    res_test = 0.5*(wavel_test[1:]+wavel_test[:-1])/np.diff(wavel_test)

    # R = lambda / delta_lambda / 2, because twice as many points as R are required to resolve
    # two features that are lambda / R apart

    wavelength = np.logspace(np.log10(wavel_range[0]),
                             np.log10(wavel_range[1]),
                             math.ceil(2.*n_test*spec_res/np.mean(res_test))+1)

    # res_out = np.mean(0.5*(wavelength[1:]+wavelength[:-1])/np.diff(wavelength)/2.)

    return wavelength


@typechecked
def smooth_spectrum(wavelength: np.ndarray,
                    flux: np.ndarray,
                    spec_res: float,
                    size: int = 11) -> np.ndarray:
    """
    Function for smoothing a spectrum with a Gaussian kernel to a fixed spectral resolution. The
    kernel size is set to 5 times the FWHM of the Gaussian. The FWHM of the Gaussian is equal
    to the ratio of the wavelength and the spectral resolution. If the kernel does not fit within
    the available wavelength grid (i.e. at the edge of the array) then the flux values are set
    to NaN.

    Parameters
    ----------
    wavelength : np.ndarray
        Wavelength points (um). Should be sampled with a uniform spectral resolution or a uniform
        wavelength spacing (slow).
    flux : np.ndarray
        Flux (W m-2 um-1).
    spec_res : float
        Spectral resolution.
    size : int
        Kernel size (odd integer).

    Returns
    -------
    np.ndarray
        Smoothed spectrum (W m-2 um-1).
    """

    def _gaussian(size, sigma):
        pos = range(-(size-1)//2, (size-1)//2+1)
        kernel = [np.exp(-float(x)**2/(2.*sigma**2))/(sigma*np.sqrt(2.*np.pi)) for x in pos]

        return np.asarray(kernel/sum(kernel))

    spacing = np.mean(2.*np.diff(wavelength)/(wavelength[1:]+wavelength[:-1]))
    spacing_std = np.std(2.*np.diff(wavelength)/(wavelength[1:]+wavelength[:-1]))

    if spacing_std/spacing < 1e-2:
        # see retrieval_util.convolve
        sigma_lsf = 1./spec_res/(2.*np.sqrt(2.*np.log(2.)))
        flux_smooth = gaussian_filter(flux, sigma=sigma_lsf/spacing, mode='nearest')

    else:
        if size % 2 == 0:
            raise ValueError('The kernel size should be an odd number.')

        flux_smooth = np.zeros(flux.shape)  # (W m-2 um-1)

        spacing = np.mean(np.diff(wavelength))  # (um)
        spacing_std = np.std(np.diff(wavelength))  # (um)

        if spacing_std/spacing > 1e-2:
            warnings.warn(f'The wavelength spacing is not uniform ({spacing} +/- {spacing_std}). '
                          f'The smoothing with the Gaussian kernel requires either the spectral '
                          f'resolution or the wavelength spacing to be uniformly sampled.')

        for i, item in enumerate(wavelength):
            fwhm = item/spec_res  # (um)
            sigma = fwhm/(2.*np.sqrt(2.*np.log(2.)))  # (um)

            size = int(5.*sigma/spacing)  # Kernel size 5 times the width of the LSF
            if size % 2 == 0:
                size += 1

            gaussian = _gaussian(size, sigma/spacing)

            try:
                flux_smooth[i] = np.sum(gaussian * flux[i-(size-1)//2:i+(size-1)//2+1])

            except ValueError:
                flux_smooth[i] = np.nan

    return flux_smooth


@typechecked
def powerlaw_spectrum(wavel_range: Union[Tuple[float, float],
                                         Tuple[np.float32, np.float32]],
                      model_param: Dict[str, float],
                      spec_res: float = 100.) -> box.ModelBox:
    """
    Function for calculating a power-law spectrum. The power-law function is calculated in
    log(wavelength)-log(flux) space but stored in the :class:`~species.core.box.ModelBox`
    in linear wavelength-flux space.

    Parameters
    ----------
    wavel_range : tuple(float, float)
        Tuple with the minimum and maximum wavelength (um).
    model_param : dict
        Dictionary with the model parameters. Should contain `'log_powerlaw_a'`,
        `'log_powerlaw_b'`, and `'log_powerlaw_c'`.
    spec_res : float
        Spectral resolution (default: 100).

    Returns
    -------
    species.core.box.ModelBox
        Box with the power-law spectrum.
    """

    wavel = create_wavelengths((wavel_range[0], wavel_range[1]), spec_res)

    log_flux = model_param['log_powerlaw_a'] + \
        model_param['log_powerlaw_b']*np.log10(wavel)**model_param['log_powerlaw_c']

    model_box = box.create_box(boxtype='model',
                               model='powerlaw',
                               wavelength=wavel,
                               flux=10.**log_flux,
                               parameters=model_param,
                               quantity='flux')

    return model_box



import os
import shutil

import pytest
import numpy as np

import species
from species.util import test_util


class TestModel:

    def setup_class(self):
        self.limit = 1e-8
        self.test_path = os.path.dirname(__file__) + '/'
        self.model_param = {'teff': 2200., 'logg': 4.5, 'radius': 1., 'distance': 10.}

    def teardown_class(self):
        os.remove('species_database.hdf5')
        os.remove('species_config.ini')
        shutil.rmtree('data/')

    def test_species_init(self):
        test_util.create_config('./')
        species.SpeciesInit()

    def test_read_model(self):
        database = species.Database()

        database.add_model('ames-cond',
                           wavel_range=(1., 5.),
                           spec_res=100.,
                           teff_range=(2000., 2500))

        read_model = species.ReadModel('ames-cond')
        assert read_model.model == 'ames-cond'

    def test_get_model(self):
        read_model = species.ReadModel('ames-cond', filter_name='Paranal/NACO.H')

        model_box = read_model.get_model(self.model_param,
                                         spec_res=100.,
                                         magnitude=False,
                                         smooth=True)

        assert np.sum(model_box.wavelength) == pytest.approx(92.26773310928259, rel=self.limit, abs=0.)
        assert np.sum(model_box.flux) == pytest.approx(1.6347074150483604e-12, rel=self.limit, abs=0.)

        model_box = read_model.get_model(self.model_param,
                                         spec_res=100.,
                                         magnitude=True,
                                         smooth=True)

        assert np.sum(model_box.wavelength) == pytest.approx(92.26773310928259, rel=self.limit, abs=0.)
        assert np.sum(model_box.flux) == pytest.approx(650.0527540140317, rel=self.limit, abs=0.)

    def test_get_data(self):
        read_model = species.ReadModel('ames-cond', filter_name='Paranal/NACO.H')
        model_box = read_model.get_data(self.model_param)

        assert np.sum(model_box.wavelength) == pytest.approx(92.26773310928259, rel=self.limit, abs=0.)
        assert np.sum(model_box.flux) == pytest.approx(1.6346900092886714e-12, rel=self.limit, abs=0.)

    def test_get_flux(self):
        read_model = species.ReadModel('ames-cond', filter_name='Paranal/NACO.H')
        flux = read_model.get_flux(self.model_param)

        assert flux[0] == pytest.approx(3.336861738800749e-14, rel=self.limit, abs=0.)

    def test_get_magnitude(self):
        read_model = species.ReadModel('ames-cond', filter_name='Paranal/NACO.H')
        magnitude = read_model.get_magnitude(self.model_param)

        assert magnitude[0] == pytest.approx(11.357135619661243, rel=self.limit, abs=0.)
        assert magnitude[1] == pytest.approx(11.357135619661243, rel=self.limit, abs=0.)

    def test_get_bounds(self):
        read_model = species.ReadModel('ames-cond', filter_name='Paranal/NACO.H')
        bounds = read_model.get_bounds()

        assert bounds['teff'] == (2000., 2500.)
        assert bounds['logg'] == (2.5, 5.5)

    def test_get_wavelengths(self):
        read_model = species.ReadModel('ames-cond', filter_name='Paranal/NACO.H')
        wavelengths = read_model.get_wavelengths()

        assert np.sum(wavelengths) == pytest.approx(813.2224003071026, rel=1e-7, abs=0.)

    def test_get_points(self):
        read_model = species.ReadModel('ames-cond', filter_name='Paranal/NACO.H')
        points = read_model.get_points()

        assert np.sum(points['teff']) == 13500.
        assert np.sum(points['logg']) == 28.

    def test_get_parameters(self):
        read_model = species.ReadModel('ames-cond', filter_name='Paranal/NACO.H')
        parameters = read_model.get_parameters()

        assert parameters == ['teff', 'logg']



"""
Module for Exo-REM atmospheric model spectra.
"""

import os
import tarfile
import urllib.request

from typing import Optional, Tuple

import h5py
import spectres
import numpy as np

from typeguard import typechecked

from species.util import data_util, read_util


@typechecked
def add_exo_rem(input_path: str,
                database: h5py._hl.files.File,
                wavel_range: Optional[Tuple[float, float]] = None,
                teff_range: Optional[Tuple[float, float]] = None,
                spec_res: Optional[float] = None) -> None:
    """
    Function for adding the Exo-REM atmospheric models to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    wavel_range : tuple(float, float), None
        Wavelength range (um). The original wavelength points with a spectral resolution of 5000
        are used if set to ``None``.
    teff_range : tuple(float, float), None
        Effective temperature range (K). All temperatures are selected if set to ``None``.
    spec_res : float, None
        Spectral resolution. Not used if ``wavel_range`` is set to ``None``.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    input_file = 'exo-rem.tgz'
    url = 'https://people.phys.ethz.ch/~ipa/tstolker/exo-rem.tgz'

    data_folder = os.path.join(input_path, 'exo-rem/')
    data_file = os.path.join(input_path, input_file)

    if not os.path.exists(data_folder):
        os.makedirs(data_folder)

    if not os.path.isfile(data_file):
        print('Downloading Exo-REM model spectra (706 MB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Unpacking Exo-REM model spectra (706 MB)...', end='', flush=True)
    tar = tarfile.open(data_file)
    tar.extractall(data_folder)
    tar.close()
    print(' [DONE]')

    teff = []
    logg = []
    feh = []
    co_ratio = []
    flux = []

    if wavel_range is not None and spec_res is not None:
        wavelength = read_util.create_wavelengths(wavel_range, spec_res)
    else:
        wavelength = None

    for _, _, files in os.walk(data_folder):
        for filename in files:
            if filename[:8] == 'exo-rem_':
                file_split = filename.split('_')

                teff_val = float(file_split[2])
                logg_val = float(file_split[4])
                feh_val = float(file_split[6])
                co_val = float(file_split[8])

                if logg_val == 5.:
                    continue

                if co_val in [0.8, 0.85]:
                    continue

                if teff_range is not None:
                    if teff_val < teff_range[0] or teff_val > teff_range[1]:
                        continue

                print_message = f'Adding Exo-REM model spectra... {filename}'
                print(f'\r{print_message:<84}', end='')

                data_wavel, data_flux = np.loadtxt(os.path.join(data_folder, filename), unpack=True)

                teff.append(teff_val)
                logg.append(logg_val)
                feh.append(feh_val)
                co_ratio.append(co_val)

                if wavel_range is None or spec_res is None:
                    if wavelength is None:
                        wavelength = np.copy(data_wavel)  # (um)

                    if np.all(np.diff(wavelength) < 0):
                        raise ValueError('The wavelengths are not all sorted by increasing value.')

                    flux.append(data_flux)  # (W m-2 um-1)

                else:
                    flux_resample = spectres.spectres(wavelength,
                                                      data_wavel,
                                                      data_flux,
                                                      spec_errs=None,
                                                      fill=np.nan,
                                                      verbose=False)

                    if np.isnan(np.sum(flux_resample)):
                        raise ValueError(f'Resampling is only possible if the new wavelength '
                                         f'range ({wavelength[0]} - {wavelength[-1]} um) falls '
                                         f'sufficiently far within the wavelength range '
                                         f'({data_wavel[0]} - {data_wavel[-1]} um) of the input '
                                         f'spectra.')

                    flux.append(flux_resample)  # (W m-2 um-1)

    print_message = 'Adding Exo-REM model spectra... [DONE]'
    print(f'\r{print_message:<84}')

    print('Grid points with the following parameters have been excluded:')
    print('   - log(g) = 5')
    print('   - C/O = 0.8')
    print('   - C/O = 0.85')

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      np.asarray(co_ratio),
                                      None,
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('exo-rem',
                         ['teff', 'logg', 'feh', 'co'],
                         database,
                         data_sorted)



"""
Module with reading functionalities for filter profiles.
"""

import os
import warnings
import configparser

from typing import Union, Tuple

import h5py
import numpy as np

from typeguard import typechecked
from scipy.interpolate import interp1d, InterpolatedUnivariateSpline, interpolate

from species.data import database


class ReadFilter:
    """
    Class for reading a filter profile from the database.
    """

    @typechecked
    def __init__(self,
                 filter_name: str) -> None:
        """
        Parameters
        ----------
        filter_name : str
            Filter name as stored in the database. Filter names from the SVO Filter Profile Service
            will be automatically downloaded, stored in the database, and read from the database.

        Returns
        -------
        NoneType
            None
        """

        self.filter_name = filter_name

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

        h5_file = h5py.File(self.database, 'r')

        if 'filters' not in h5_file or self.filter_name not in h5_file['filters']:
            h5_file.close()
            species_db = database.Database()
            species_db.add_filter(self.filter_name)
            h5_file = h5py.File(self.database, 'r')

        h5_file.close()

    @typechecked
    def get_filter(self) -> np.ndarray:
        """
        Function for selecting a filter profile from the database.

        Returns
        -------
        np.ndarray
            Array with the wavelengths and filter transmission.
        """

        h5_file = h5py.File(self.database, 'r')

        data = np.asarray(h5_file[f'filters/{self.filter_name}'])

        if data.shape[0] == 2 and data.shape[1] > data.shape[0]:
            # Required for backward compatibility
            data = np.transpose(data)

        h5_file.close()

        return data

    @typechecked
    def interpolate_filter(self) -> interpolate.interp1d:
        """
        Function for linearly interpolating a filter profile.

        Returns
        -------
        scipy.interpolate.interpolate.interp1d
            Linearly interpolated filter.
        """

        data = self.get_filter()

        return interp1d(data[:, 0],
                        data[:, 1],
                        kind='linear',
                        bounds_error=False,
                        fill_value=float('nan'))

    @typechecked
    def wavelength_range(self) -> Tuple[Union[np.float32, np.float64],
                                        Union[np.float32, np.float64]]:
        """
        Extract the wavelength range of the filter profile.

        Returns
        -------
        float
            Minimum wavelength (um).
        float
            Maximum wavelength (um).
        """

        data = self.get_filter()

        return data[0, 0], data[-1, 0]

    @typechecked
    def mean_wavelength(self) -> Union[np.float32, np.float64]:
        """
        Calculate the weighted mean wavelength of the filter profile.

        Returns
        -------
        float
            Mean wavelength (um).
        """

        data = self.get_filter()

        return np.trapz(data[:, 0]*data[:, 1], data[:, 0]) / np.trapz(data[:, 1], data[:, 0])

    @typechecked
    def filter_fwhm(self) -> float:
        """
        Calculate the full width at half maximum (FWHM) of the filter profile.

        Returns
        -------
        float
            Filter full width at half maximum (um).
        """

        data = self.get_filter()

        spline = InterpolatedUnivariateSpline(data[:, 0], data[:, 1] - np.max(data[:, 1])/2.)
        root = spline.roots()

        diff = root - self.mean_wavelength()

        root1 = np.amax(diff[diff < 0.])
        root2 = np.amin(diff[diff > 0.])

        return root2 - root1

    @typechecked
    def detector_type(self) -> str:
        """
        Function for returning the detector type.

        Returns
        -------
        str
            Detector type ('energy' or 'photon').
        """

        with h5py.File(self.database, 'r') as h5_file:
            dset = h5_file[f'filters/{self.filter_name}']

            if 'det_type' in dset.attrs:
                det_type = dset.attrs['det_type']

            else:
                warnings.warn('Detector type not found. The database was probably created '
                              'before the detector type was introduced in species (v0.3.1). '
                              'Assuming an energy-counting detector.')

                det_type = 'energy'

        return det_type



import os
import shutil

import pytest
import numpy as np

import species
from species.util import test_util


class TestColor:

    def setup_class(self):
        self.limit = 1e-8

    def teardown_class(self):
        os.remove('species_database.hdf5')
        os.remove('species_config.ini')
        shutil.rmtree('data/')

    def test_species_init(self):
        test_util.create_config('./')
        species.SpeciesInit()

    def test_read_color_magnitude(self):
        database = species.Database()
        database.add_photometry('vlm-plx')
        database.add_photometry('leggett')

        read_colormag = species.ReadColorMagnitude('vlm-plx',
                                                   ('MKO/NSFCam.J', 'MKO/NSFCam.H'),
                                                   'MKO/NSFCam.J')

        assert read_colormag.filters_color == ('MKO/NSFCam.J', 'MKO/NSFCam.H')
        assert read_colormag.filter_mag == 'MKO/NSFCam.J'

    def test_get_color_magnitude(self):
        read_colormag = species.ReadColorMagnitude('leggett',
                                                   ('MKO/NSFCam.J', 'MKO/NSFCam.H'),
                                                   'MKO/NSFCam.J')

        colormag_box = read_colormag.get_color_magnitude(object_type=None)
        colormag_box.open_box()

        assert np.nansum(colormag_box.color) == pytest.approx(14.05)
        assert np.nansum(colormag_box.magnitude) == pytest.approx(1519.373)

    def test_read_color_color(self):
        read_colorcolor = species.ReadColorColor('vlm-plx',
                                                 (('MKO/NSFCam.J', 'MKO/NSFCam.H'),
                                                  ('MKO/NSFCam.H', 'MKO/NSFCam.K')))

        assert read_colorcolor.filters_colors == (('MKO/NSFCam.J', 'MKO/NSFCam.H'),
                                                  ('MKO/NSFCam.H', 'MKO/NSFCam.K'))

    def test_get_color_color(self):
        read_colorcolor = species.ReadColorColor('vlm-plx',
                                                 (('MKO/NSFCam.J', 'MKO/NSFCam.H'),
                                                  ('MKO/NSFCam.H', 'MKO/NSFCam.K')))

        colorcolor_box = read_colorcolor.get_color_color(object_type=None)

        assert np.nansum(colorcolor_box.color1) == pytest.approx(166.99384)
        assert np.nansum(colorcolor_box.color2) == pytest.approx(138.5127)



"""
Module for setting up species in the working folder.
"""

import os
import json
import configparser
import urllib.request

import h5py
import species


class SpeciesInit:
    """
    Class for initiating species by creating the database and configuration file in case they are
    not present in the working folder, and creating the data folder for storage of input data.
    """

    def __init__(self):
        """
        Returns
        -------
        NoneType
            None
        """

        print(f'Initiating species v{species.__version__}...', end='', flush=True)

        working_folder = os.path.abspath(os.getcwd())

        config_file = os.path.join(working_folder, 'species_config.ini')

        print(' [DONE]')

        try:
            contents = urllib.request.urlopen('https://pypi.org/pypi/species/json',
                                              timeout=1.).read()

            data = json.loads(contents)
            latest_version = data['info']['version']

        except urllib.error.URLError:
            latest_version = None

        if latest_version is not None and species.__version__ != latest_version:
            print(f'A new version ({latest_version}) is available!')
            print('Want to stay informed about updates, bug fixes, and new features?')
            print('Please consider using \'Watch\' button on the Github page:')
            print('https://github.com/tomasstolker/species')

        if not os.path.isfile(config_file):

            print('Creating species_config.ini...', end='', flush=True)

            with open(config_file, 'w') as file_obj:
                file_obj.write('[species]\n')
                file_obj.write('database = species_database.hdf5\n')
                file_obj.write('data_folder = ./data/\n')

            print(' [DONE]')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        database_file = os.path.abspath(config['species']['database'])
        data_folder = os.path.abspath(config['species']['data_folder'])

        print(f'Database: {database_file}')
        print(f'Data folder: {data_folder}')
        print(f'Working folder: {working_folder}')

        if not os.path.isfile(database_file):
            print('Creating species_database.hdf5...', end='', flush=True)
            h5_file = h5py.File(database_file, 'w')
            h5_file.close()
            print(' [DONE]')

        if not os.path.exists(data_folder):
            print('Creating data folder...', end='', flush=True)
            os.makedirs(data_folder)
            print(' [DONE]')



"""
Module with functionalities for fitting atmospheric model spectra.
"""

import os
import math
import warnings

from typing import Optional, Union, List, Tuple, Dict
from multiprocessing import Pool, cpu_count

import emcee
import spectres
import numpy as np

# Installation of MultiNest is not possible on readthedocs
try:
    import pymultinest
except:
    warnings.warn('PyMultiNest could not be imported.')

from typeguard import typechecked

from species.analysis import photometry
from species.data import database
from species.core import constants
from species.read import read_model, read_object, read_planck, read_filter
from species.util import read_util, dust_util


@typechecked
def lnprior(param: np.ndarray,
            bounds: dict,
            param_index: Dict[str, int],
            prior: Optional[Dict[str, Tuple[float, float]]] = None):
    """
    Internal function for calculating the log prior.

    Parameters
    ----------
    param : np.ndarray
        Parameter values.
    bounds : dict
        Dictionary with the parameter boundaries.
    param_index : dict(str, int)
        Dictionary with the parameter indices of ``param``.
    prior : dict(str, tuple(float, float)), None
        Dictionary with Gaussian priors for one or multiple parameters. The prior can be set
        for any of the atmosphere or calibration parameters, e.g.
        ``prior={'teff': (1200., 100.)}``. Additionally, a prior can be set for the mass, e.g.
        ``prior={'mass': (13., 3.)}`` for an expected mass of 13 Mjup with an uncertainty of
        3 Mjup. The parameter is not used if set to ``None``.

    Returns
    -------
    floats
        Log prior.
    """

    ln_prior = 0

    for key, value in bounds.items():

        if value[0] <= param[param_index[key]] <= value[1]:
            ln_prior += 0.

        else:
            ln_prior = -np.inf
            break

    if prior is not None:
        for key, value in prior.items():
            if key == 'mass':
                mass = read_util.get_mass(param[param_index['logg']],
                                          param[param_index['radius']])

                ln_prior += -0.5 * (mass - value[0])**2 / value[1]**2

            else:
                ln_prior += -0.5 * (param[param_index[key]] - value[0])**2 / value[1]**2

    return ln_prior


@typechecked
def lnlike(param: np.ndarray,
           bounds: dict,
           param_index: Dict[str, int],
           model: str,
           objphot: List[Optional[np.ndarray]],
           distance: Tuple[float, float],
           spectrum: Optional[dict],
           modelphot: Optional[Union[List[read_model.ReadModel],
                                     List[photometry.SyntheticPhotometry]]],
           modelspec: Optional[List[read_model.ReadModel]],
           n_planck: int,
           fit_corr: List[str]):
    """
    Internal function for calculating the log likelihood.

    Parameters
    ----------
    param : np.ndarray
        Parameter values.
    bounds : dict
        Dictionary with the parameter boundaries.
    param_index : dict(str, int)
        Dictionary with the parameter indices of ``param``.
    model : str
        Atmosphere model (e.g. 'bt-settl', 'exo-rem', or 'planck).
    objphot : list(np.ndarray, )
        List with the photometric fluxes and uncertainties of the object. Not photometric data
        is fitted if an empty list is provided.
    distance : tuple(float, float)
        Distance and uncertainty (pc).
    spectrum : dict(str, tuple(np.ndarray, np.ndarray, np.ndarray, float)), None
        Dictionary with the spectra stored as wavelength (um), flux (W m-2 um-1),
        and error (W m-2 um-1). Optionally the covariance matrix, the inverse of
        the covariance matrix, and the spectral resolution are included. Each
        of these three elements can be set to ``None``. No spectroscopic data is
        fitted if ``spectrum=None``.
    modelphot : list(species.read.read_model.ReadModel, ),
                list(species.analysis.photometry.SyntheticPhotometry, ), None
        List with the interpolated synthetic fluxes or list with the
        :class:`~species.analysis.photometry.SyntheticPhotometry` objects for
        calculation of synthetic photometry for Planck spectra. No photometry is fitted if set
        to ``None``.
    modelspec : list(species.read.read_model.ReadModel, ), None
        List with the interpolated synthetic spectra.
    n_planck : int
        Number of Planck components. The argument is set to zero if ``model`` is not ``'planck'``.
    fit_corr : list(str)
        List with spectrum names for which the correlation length and fractional amplitude are
        fitted (see Wang et al. 2020).

    Returns
    -------
    float
        Log likelihood.
    """

    param_dict = {}
    spec_scaling = {}
    err_scaling = {}
    corr_len = {}
    corr_amp = {}

    for item in bounds:
        if item[:8] == 'scaling_' and item[8:] in spectrum:
            spec_scaling[item[8:]] = param[param_index[item]]

        elif item[:6] == 'error_' and item[6:] in spectrum:
            err_scaling[item[6:]] = param[param_index[item]]

        elif item[:9] == 'corr_len_' and item[9:] in spectrum:
            corr_len[item[9:]] = 10.**param[param_index[item]]  # (um)

        elif item[:9] == 'corr_amp_' and item[9:] in spectrum:
            corr_amp[item[9:]] = param[param_index[item]]

        else:
            param_dict[item] = param[param_index[item]]

    if model == 'planck':
        param_dict['distance'] = distance[0]

    else:
        flux_scaling = (param_dict['radius']*constants.R_JUP)**2 / \
                       (distance[0]*constants.PARSEC)**2

        # The scaling is applied manually because of the interpolation
        del param_dict['radius']

    for item in spectrum:
        if item not in spec_scaling:
            spec_scaling[item] = 1.

        if item not in err_scaling:
            err_scaling[item] = None

    ln_like = 0.

    if model == 'planck' and n_planck > 1:
        for i in range(n_planck-1):
            if param_dict[f'teff_{i+1}'] > param_dict[f'teff_{i}']:
                return -np.inf

            if param_dict[f'radius_{i}'] > param_dict[f'radius_{i+1}']:
                return -np.inf

    for i, obj_item in enumerate(objphot):
        if model == 'planck':
            readplanck = read_planck.ReadPlanck(filter_name=modelphot[i].filter_name)
            phot_flux = readplanck.get_flux(param_dict, synphot=modelphot[i])[0]

        else:
            phot_flux = modelphot[i].spectrum_interp(list(param_dict.values()))
            phot_flux *= flux_scaling

        if obj_item.ndim == 1:
            ln_like += -0.5 * (obj_item[0] - phot_flux)**2 / obj_item[1]**2

        else:
            for j in range(obj_item.shape[1]):
                ln_like += -0.5 * (obj_item[0, j] - phot_flux)**2 / obj_item[1, j]**2

    for i, item in enumerate(spectrum.keys()):
        # Calculate or interpolate the model spectrum
        if model == 'planck':
            # Calculate a blackbody spectrum
            readplanck = read_planck.ReadPlanck((0.9*spectrum[item][0][0, 0],
                                                 1.1*spectrum[item][0][-1, 0]))

            model_box = readplanck.get_spectrum(param_dict, 1000., smooth=True)

            # Resample the spectrum to the observed wavelengths
            model_flux = spectres.spectres(spectrum[item][0][:, 0],
                                           model_box.wavelength,
                                           model_box.flux)

        else:
            # Interpolate the model spectrum
            model_flux = modelspec[i].spectrum_interp(list(param_dict.values()))[0, :]

            # Scale the spectrum by the (radius/distance)^2
            model_flux *= flux_scaling

        # Scale the spectrum data
        data_flux = spec_scaling[item]*spectrum[item][0][:, 1]

        if err_scaling[item] is None:
            # Variance without error inflation
            data_var = spectrum[item][0][:, 2]**2

        else:
            # Variance with error inflation (see Piette & Madhusudhan 2020)
            data_var = spectrum[item][0][:, 2]**2 + (err_scaling[item]*model_flux)**2

        if spectrum[item][2] is not None:
            # The inverted covariance matrix is available

            if err_scaling[item] is None:
                # Use the inverted covariance matrix directly
                data_cov_inv = spectrum[item][2]

            else:
                # Ratio of the inflated and original uncertainties
                sigma_ratio = np.sqrt(data_var) / spectrum[item][0][:, 2]
                sigma_j, sigma_i = np.meshgrid(sigma_ratio, sigma_ratio)

                # Calculate the inverted matrix of the inflated covariances
                data_cov_inv = np.linalg.inv(spectrum[item][1]*sigma_i*sigma_j)

        if spectrum[item][2] is not None:
            # Calculate the log-likelihood with the covariance matrix
            dot_tmp = np.dot(data_flux-model_flux, np.dot(data_cov_inv, data_flux-model_flux))
            ln_like += -0.5*dot_tmp - 0.5*np.nansum(np.log(2.*np.pi*data_var))

        else:
            if item in fit_corr:
                # Calculate the log-likelihood with the covariance model (see Wang et al. 2020)
                wavel = spectrum[item][0][:, 0]  # (um)
                wavel_j, wavel_i = np.meshgrid(wavel, wavel)

                error = np.sqrt(data_var)  # (W m-2 um-1)
                error_j, error_i = np.meshgrid(error, error)

                cov_matrix = corr_amp[item]**2 * error_i * error_j * \
                    np.exp(-(wavel_i-wavel_j)**2 / (2.*corr_len[item]**2)) + \
                    (1.-corr_amp[item]**2) * np.eye(wavel.shape[0])*error_i**2

                dot_tmp = np.dot(data_flux-model_flux,
                                 np.dot(np.linalg.inv(cov_matrix), data_flux-model_flux))

                ln_like += -0.5*dot_tmp - 0.5*np.nansum(np.log(2.*np.pi*data_var))

            else:
                # Calculate the log-likelihood without the covariance matrix but with the
                # normalization term in case of the errors are inflated
                ln_like += np.nansum(-0.5 * (data_flux-model_flux)**2 / data_var -
                                     0.5 * np.log(2.*np.pi*data_var))

    return ln_like


@typechecked
def lnprob(param: np.ndarray,
           bounds: dict,
           model: str,
           param_index: Dict[str, int],
           objphot: List[Optional[np.ndarray]],
           distance: Tuple[float, float],
           prior: Optional[Dict[str, Tuple[float, float]]],
           spectrum: Optional[dict],
           modelphot: Optional[Union[List[read_model.ReadModel],
                                     List[photometry.SyntheticPhotometry]]],
           modelspec: Optional[List[read_model.ReadModel]],
           n_planck: int,
           fit_corr: List[str]) -> np.float64:
    """
    Internal function for calculating the log posterior.

    Parameters
    ----------
    param : np.ndarray
        Parameter values.
    bounds : dict
        Parameter boundaries.
    model : str
        Atmosphere model (e.g. 'bt-settl', 'exo-rem', or 'planck).
    param_index : dict(str, int)
        Dictionary with the parameter indices of ``param``.
    objphot : list(np.ndarray, ), None
        List with the photometric fluxes and uncertainties. No photometric data is fitted if the
        parameter is set to ``None``.
    distance : tuple(float, float)
        Distance and uncertainty (pc).
    prior : dict(str, tuple(float, float)), None
        Dictionary with Gaussian priors for one or multiple parameters. The prior can be set
        for any of the atmosphere or calibration parameters, e.g.
        ``prior={'teff': (1200., 100.)}``. Additionally, a prior can be set for the mass, e.g.
        ``prior={'mass': (13., 3.)}`` for an expected mass of 13 Mjup with an uncertainty of
        3 Mjup. The parameter is not used if set to ``None``.
    spectrum : dict(str, tuple(np.ndarray, np.ndarray, np.ndarray, float)), None
        Dictionary with the spectra stored as wavelength (um), flux (W m-2 um-1),
        and error (W m-2 um-1). Optionally the covariance matrix, the inverse of
        the covariance matrix, and the spectral resolution are included. Each
        of these three elements can be set to ``None``. No spectroscopic data is
        fitted if ``spectrum=None``.
    modelphot : list(species.read.read_model.ReadModel, ),
                list(species.analysis.photometry.SyntheticPhotometry, ), None
        List with the interpolated synthetic fluxes or list with the
        :class:`~species.analysis.photometry.SyntheticPhotometry` objects for
        calculation of synthetic photometry for Planck spectra. No photometry is fitted if set
        to ``None``.
    modelspec : list(species.read.read_model.ReadModel, ), None
        List with the interpolated synthetic spectra. The parameter is set to ``None`` when no
        spectroscopic data is included or when ``model='planck'``.
    n_planck : int
        Number of Planck components. The argument is set to zero if ``model`` is not ``'planck'``.
    fit_corr : list(str)
        List with spectrum names for which the correlation amplitude and length are fitted.

    Returns
    -------
    float
        Log posterior.
    """

    ln_prior = lnprior(param, bounds, param_index, prior)

    if math.isinf(ln_prior):
        ln_prob = -np.inf

    else:
        ln_prob = ln_prior + lnlike(param, bounds, param_index, model, objphot, distance, spectrum,
                                    modelphot, modelspec, n_planck, fit_corr)

    if np.isnan(ln_prob):
        ln_prob = -np.inf

    return ln_prob


class FitModel:
    """
    Class for fitting atmospheric model spectra or blackbody spectra to photometric and/or
    spectroscopic data.
    """

    @typechecked
    def __init__(self,
                 object_name: str,
                 model: str,
                 bounds: Dict[str, Union[Tuple[float, float],
                                         Tuple[Optional[Tuple[float, float]],
                                               Optional[Tuple[float, float]]],
                                         List[Tuple[float, float]]]],
                 inc_phot: Union[bool, List[str]] = True,
                 inc_spec: Union[bool, List[str]] = True,
                 fit_corr: Optional[List[str]] = None) -> None:
        """
        The grid of spectra is linearly interpolated for each photometric point and spectrum while
        taking into account the filter profile, spectral resolution, and wavelength sampling.
        Therefore, when fitting spectra from a model grid, the computation time of the
        interpolation will depend on the wavelength range, spectral resolution, and parameter
        space of the spectra that are stored in the database.

        Parameters
        ----------
        object_name : str
            Object name as stored in the database with
            :func:`~species.data.database.Database.add_object` or
            :func:`~species.data.database.Database.add_companion`.
        model : str
            Atmospheric model (e.g. 'bt-settl', 'exo-rem', 'planck', or 'powerlaw').
        bounds : dict(str, tuple(float, float)), None
            The boundaries that are used for the uniform priors. Fixing a parameter is possible by
            providing the same value as lower and upper boundary of the parameter, for example,
            ``bounds={'logg': (4., 4.)``.

            Atmospheric model parameters (e.g. ``model='bt-settl'``):

                 - Boundaries are provided as tuple of two floats. For example,
                   ``bounds={'teff': (1000, 1500.), 'logg': (3.5, 5.), 'radius': (0.8, 1.2)}``.

                 - The grid boundaries are used if set to ``None``. For example,
                   ``bounds={'teff': None, 'logg': None}``. The radius range is set to
                   0.8-1.5 Rjup if the boundary is set to None.

            Blackbody emission parameters (``model='planck'``):

                 - Parameter boundaries have to be provided for 'teff' and 'radius'.

                 - For a single blackbody component, the values are provided as a tuple with two
                   floats. For example, ``bounds={'teff': (1000., 2000.), 'radius': (0.8, 1.2)}``.

                 - For multiple blackbody component, the values are provided as a list with tuples.
                   For example, ``bounds={'teff': [(1000., 1400.), (1200., 1600.)],
                   'radius': [(0.8, 1.5), (1.2, 2.)]}``.

                 - When fitting multiple blackbody components, a prior is used which restricts the
                   temperatures and radii to decreasing and increasing values, respectively, in the
                   order as provided in ``bounds``.

            Power-law spectrum (``model='powerlaw'``):

                 - Parameter boundaries have to be provided for 'log_powerlaw_a', 'log_powerlaw_b',
                   and 'log_powerlaw_c'. For example, ``bounds={'log_powerlaw_a': (-20., 0.),
                   'log_powerlaw_b': (-20., 5.), 'log_powerlaw_c': (-20., 5.)}``.

                 - The spectrum is parametrized as log10(flux) = a + b*log10(wavelength)^c, where
                   a = log_powerlaw_a, b = log_powerlaw_b, and c = log_powerlaw_c.

                 - Only implemented for fitting photometric fluxes, for example the IR fluxes of a
                   star with disk. In that way, synthetic photometry can be calculated afterwards
                   for a different filter. Note that this option assumes that the photometric
                   fluxes are dominated by continuum emission while spectral lines are ignored.

                 - The :func:`~species.plot.plot_mcmc.plot_photometry` function can be used to
                   calculate synthetic photometry and error bars from the posterior distribution.

                 - Only supported by ``run_multinest``.

            Calibration parameters:

                 - For each spectrum/instrument, two optional parameters can be fitted to account
                   for biases in the calibration: a scaling of the flux and a constant inflation of
                   the uncertainties.

                 - For example, ``bounds={'SPHERE': ((0.8, 1.2), (-18., -14.))}`` if the scaling is
                   fitted between 0.8 and 1.2, and the error is inflated with a value between 1e-18
                   and 1e-14 W m-2 um-1.

                 - The dictionary key should be equal to the database tag of the spectrum. For
                   example, ``{'SPHERE': ((0.8, 1.2), (-18., -14.))}`` if the spectrum is stored as
                   ``'SPHERE'`` with :func:`~species.data.database.Database.add_object`.

                 - Each of the two calibration parameters can be set to ``None`` in which case the
                   parameter is not used. For example, ``bounds={'SPHERE': ((0.8, 1.2), None)}``.

                 - The errors of the photometric fluxes can be inflated to account for
                   underestimated error bars. The error inflation is relative to the actual flux
                   and is fitted separately for different filters. The keyword in the ``bounds``
                   dictionary should be provided in the following format:
                   ``'Paranal/NACO.Mp_error': (0., 1.)``. In this case, the error of the NACO Mp
                   flux is inflated up to 100 percent of the actual flux.

                 - No calibration parameters are fitted if the spectrum name is not included in
                   ``bounds``.

            ISM extinction parameters:

                 - There are three approaches for fitting extinction. The first is with the
                   empirical relation from Cardelli et al. (1989) for ISM extinction.

                 - The extinction is parametrized by the V band extinction, A_V (``ism_ext``), and
                   optionally the reddening, R_V (``ism_red``). If ``ism_red`` is not provided,
                   its value is fixed to 3.1 and not fitted.

                 - The prior boundaries of ``ism_ext`` and ``ext_red`` should be provided in the
                   ``bounds`` dictionary, for example ``bounds={'ism_ext': (0., 10.),
                   'ism_red': (0., 20.)}``.

                 - Only supported by ``run_multinest``.

            Log-normal size distribution:

                 - The second approach is fitting the extinction of a log-normal size distribution
                   of grains with a crystalline MgSiO3 composition, and a homogeneous, spherical
                   structure.

                 - The size distribution is parameterized with a mean geometric radius
                   (``lognorm_radius`` in um) and a geometric standard deviation
                   (``lognorm_sigma``, dimensionless). The grid of cross sections has been
                   calculated for mean geometric radii between 0.001 and 10 um, and geometric
                   standard deviations between 1.1 and 10.

                 - The extinction (``lognorm_ext``) is fitted in the V band (A_V in mag) and the
                   wavelength-dependent extinction cross sections are interpolated from a
                   pre-tabulated grid.

                 - The prior boundaries of ``lognorm_radius``, ``lognorm_sigma``, and
                   ``lognorm_ext`` should be provided in the ``bounds`` dictionary, for example
                   ``bounds={'lognorm_radius': (0.001, 10.), 'lognorm_sigma': (1.1, 10.),
                   'lognorm_ext': (0., 5.)}``.

                 - A uniform prior is used for ``lognorm_sigma`` and ``lognorm_ext``, and a
                   log-uniform prior for ``lognorm_radius``.

                 - Only supported by ``run_multinest``.

            Power-law size distribution:

                 - The third approach is fitting the extinction of a power-law size distribution
                   of grains, again with a crystalline MgSiO3 composition, and a homogeneous,
                   spherical structure.

                 - The size distribution is parameterized with a maximum radius (``powerlaw_max``
                   in um) and a power-law exponent (``powerlaw_exp``, dimensionless). The
                   minimum radius is fixed to 1 nm. The grid of cross sections has been calculated
                   for maximum radii between 0.01 and 100 um, and power-law exponents between -10
                   and 10.

                 - The extinction (``powerlaw_ext``) is fitted in the V band (A_V in mag) and the
                   wavelength-dependent extinction cross sections are interpolated from a
                   pre-tabulated grid.

                 - The prior boundaries of ``powerlaw_max``, ``powerlaw_exp``, and ``powerlaw_ext``
                   should be provided in the ``bounds`` dictionary, for example ``{'powerlaw_max':
                   (0.01, 100.), 'powerlaw_exp': (-10., 10.), 'powerlaw_ext': (0., 5.)}``.

                 - A uniform prior is used for ``powerlaw_exp`` and ``powerlaw_ext``, and a
                   log-uniform prior for ``powerlaw_max``.

                 - Only supported by ``run_multinest``.

            Blackbody disk emission:

                 - Additional blackbody emission can be added to the atmospheric spectrum to
                   account for thermal emission from a disk.

                 - Parameter boundaries have to be provided for 'disk_teff' and 'disk_radius'.
                   For example, ``bounds={'teff': (2000., 3000.), 'radius': (1., 5.),
                   'logg': (3.5, 4.5), 'disk_teff': (100., 2000.), 'disk_radius': (1., 100.)}``.

                 - Only supported by ``run_multinest``.

        inc_phot : bool, list(str)
            Include photometric data in the fit. If a boolean, either all (``True``) or none
            (``False``) of the data are selected. If a list, a subset of filter names (as stored in
            the database) can be provided.
        inc_spec : bool, list(str)
            Include spectroscopic data in the fit. If a boolean, either all (``True``) or none
            (``False``) of the data are selected. If a list, a subset of spectrum names (as stored
            in the database with :func:`~species.data.database.Database.add_object`) can be
            provided.
        fit_corr : list(str), None
            List with spectrum names for which the correlation length and fractional amplitude are
            fitted (see Wang et al. 2020).

        Returns
        -------
        NoneType
            None
        """

        if not inc_phot and not inc_spec:
            raise ValueError('No photometric or spectroscopic data has been selected.')

        if model == 'planck' and ('teff' not in bounds or 'radius' not in bounds):
            raise ValueError('The \'bounds\' dictionary should contain \'teff\' and \'radius\'.')

        self.object = read_object.ReadObject(object_name)
        self.distance = self.object.get_distance()

        if fit_corr is None:
            self.fit_corr = []
        else:
            self.fit_corr = fit_corr

        self.model = model
        self.bounds = bounds

        if self.model == 'planck':
            # Fitting blackbody radiation
            if isinstance(bounds['teff'], list) and isinstance(bounds['radius'], list):
                # Update temperature and radius parameters in case of multiple blackbody components
                self.n_planck = len(bounds['teff'])

                self.modelpar = []
                self.bounds = {}

                for i, item in enumerate(bounds['teff']):
                    self.modelpar.append(f'teff_{i}')
                    self.modelpar.append(f'radius_{i}')

                    self.bounds[f'teff_{i}'] = bounds['teff'][i]
                    self.bounds[f'radius_{i}'] = bounds['radius'][i]

            else:
                # Fitting a single blackbody compoentn
                self.n_planck = 1

                self.modelpar = ['teff', 'radius']
                self.bounds = bounds

        elif self.model == 'powerlaw':
            self.n_planck = 0

            self.modelpar = ['log_powerlaw_a', 'log_powerlaw_b', 'log_powerlaw_c']

        else:
            # Fitting self-consistent atmospheric models
            if self.bounds is not None:
                readmodel = read_model.ReadModel(self.model)
                bounds_grid = readmodel.get_bounds()

                for item in bounds_grid:
                    if item not in self.bounds:
                        # Set the parameter boundaries to the grid boundaries if set to None
                        self.bounds[item] = bounds_grid[item]

            else:
                # Set all parameter boundaries to the grid boundaries
                readmodel = read_model.ReadModel(self.model, None, None)
                self.bounds = readmodel.get_bounds()

            if 'radius' not in self.bounds:
                self.bounds['radius'] = (0.8, 1.5)

            self.n_planck = 0

            self.modelpar = readmodel.get_parameters()
            self.modelpar.append('radius')

            if 'disk_teff' in self.bounds and 'disk_radius' in self.bounds:
                self.modelpar.append('disk_teff')
                self.modelpar.append('disk_radius')

        # Select filters and spectra

        if isinstance(inc_phot, bool):
            if inc_phot:
                # Select all filters if True
                species_db = database.Database()
                objectbox = species_db.get_object(object_name)
                inc_phot = objectbox.filters

            else:
                inc_phot = []

        if isinstance(inc_spec, bool):
            if inc_spec:
                # Select all filters if True
                species_db = database.Database()
                objectbox = species_db.get_object(object_name)
                inc_spec = list(objectbox.spectrum.keys())

            else:
                inc_spec = []

        if inc_spec and self.model == 'powerlaw':
            warnings.warn('The \'inc_spec\' parameter is not supported when fitting a '
                          'power-law spectrum to photometric data. The argument of '
                          '\'inc_spec\' is therefore ignored.')

            inc_spec = []

        # Include photometric data

        self.objphot = []
        self.modelphot = []

        for item in inc_phot:
            if self.model == 'planck':
                # Create SyntheticPhotometry objects when fitting a Planck function
                print(f'Creating synthetic photometry: {item}...', end='', flush=True)
                self.modelphot.append(photometry.SyntheticPhotometry(item))
                print(' [DONE]')

            elif self.model == 'powerlaw':
                # Or create SyntheticPhotometry objects when fitting a power-law function
                synphot = photometry.SyntheticPhotometry(item)

                # Set the wavelength range of the filter as attribute
                synphot.zero_point()

                self.modelphot.append(synphot)

                if f'{item}_error' in self.bounds:
                    self.modelpar.append(f'{item}_error')

            else:
                # Or interpolate the model grid for each filter
                print(f'Interpolating {item}...', end='', flush=True)
                readmodel = read_model.ReadModel(self.model, filter_name=item)
                readmodel.interpolate_grid(wavel_resample=None, smooth=False, spec_res=None)
                self.modelphot.append(readmodel)
                print(' [DONE]')

            # Store the flux and uncertainty for each filter
            obj_phot = self.object.get_photometry(item)
            self.objphot.append(np.array([obj_phot[2], obj_phot[3]]))

        # Include spectroscopic data

        if inc_spec:
            # Select all spectra
            self.spectrum = self.object.get_spectrum()

            # Select the spectrum names that are not in inc_spec
            spec_remove = []
            for item in self.spectrum:
                if item not in inc_spec:
                    spec_remove.append(item)

            # Remove the spectra that are not included in inc_spec
            for item in spec_remove:
                del self.spectrum[item]

            self.n_corr_par = 0

            for item in self.spectrum:
                if item in self.fit_corr:
                    self.modelpar.append(f'corr_len_{item}')
                    self.modelpar.append(f'corr_amp_{item}')

                    self.bounds[f'corr_len_{item}'] = (-3., 0.)  # log10(corr_len) (um)
                    self.bounds[f'corr_amp_{item}'] = (0., 1.)

                    self.n_corr_par += 2

            self.modelspec = []

            if self.model != 'planck':

                for key, value in self.spectrum.items():
                    print(f'\rInterpolating {key}...', end='', flush=True)

                    wavel_range = (0.9*value[0][0, 0], 1.1*value[0][-1, 0])

                    readmodel = read_model.ReadModel(self.model, wavel_range=wavel_range)

                    readmodel.interpolate_grid(wavel_resample=self.spectrum[key][0][:, 0],
                                               smooth=True,
                                               spec_res=self.spectrum[key][3])

                    self.modelspec.append(readmodel)

                    print(' [DONE]')

        else:
            self.spectrum = {}
            self.modelspec = None
            self.n_corr_par = 0

        # Get the parameter order if interpolate_grid is used

        if self.model != 'planck' and self.model != 'powerlaw':
            readmodel = read_model.ReadModel(self.model)
            self.param_interp = readmodel.get_parameters()

        else:
            self.param_interp = None

        # Include blackbody disk

        self.diskphot = []
        self.diskspec = []

        if 'disk_teff' in self.bounds and 'disk_radius' in self.bounds:
            for item in inc_phot:
                print(f'Interpolating {item}...', end='', flush=True)
                readmodel = read_model.ReadModel('blackbody', filter_name=item)
                readmodel.interpolate_grid(wavel_resample=None, smooth=False, spec_res=None)
                self.diskphot.append(readmodel)
                print(' [DONE]')

            for key, value in self.spectrum.items():
                print(f'\rInterpolating {key}...', end='', flush=True)

                wavel_range = (0.9*value[0][0, 0], 1.1*value[0][-1, 0])

                readmodel = read_model.ReadModel('blackbody', wavel_range=wavel_range)

                readmodel.interpolate_grid(wavel_resample=self.spectrum[key][0][:, 0],
                                           smooth=True,
                                           spec_res=self.spectrum[key][3])

                self.diskspec.append(readmodel)

                print(' [DONE]')

        for item in self.spectrum:
            if item in bounds:

                if bounds[item][0] is not None:
                    # Add the flux scaling parameter
                    self.modelpar.append(f'scaling_{item}')
                    self.bounds[f'scaling_{item}'] = (bounds[item][0][0], bounds[item][0][1])

                if bounds[item][1] is not None:
                    # Add the error offset parameters
                    self.modelpar.append(f'error_{item}')
                    self.bounds[f'error_{item}'] = (bounds[item][1][0], bounds[item][1][1])

                    if self.bounds[f'error_{item}'][0] < 0.:
                        warnings.warn(f'The lower bound of \'error_{item}\' is smaller than 0. The '
                                      f'error inflation should be given relative to the model fluxes '
                                      f'so the boundaries are typically between 0 and 1.')

                    if self.bounds[f'error_{item}'][1] < 0.:
                        warnings.warn(f'The upper bound of \'error_{item}\' is smaller than 0. The '
                                      f'error inflation should be given relative to the model fluxes '
                                      f'so the boundaries are typically between 0 and 1.')

                    if self.bounds[f'error_{item}'][0] > 1.:
                        warnings.warn(f'The lower bound of \'error_{item}\' is larger than 1. The '
                                      f'error inflation should be given relative to the model fluxes '
                                      f'so the boundaries are typically between 0 and 1.')

                    if self.bounds[f'error_{item}'][1] > 1.:
                        warnings.warn(f'The upper bound of \'error_{item}\' is larger than 1. The '
                                      f'error inflation should be given relative to the model fluxes '
                                      f'so the boundaries are typically between 0 and 1.')

                if item in self.bounds:
                    del self.bounds[item]

        if 'lognorm_radius' in self.bounds and 'lognorm_sigma' in self.bounds and \
                'lognorm_ext' in self.bounds:

            self.cross_sections, _, _ = dust_util.interp_lognorm(inc_phot, inc_spec, self.spectrum)

            self.modelpar.append('lognorm_radius')
            self.modelpar.append('lognorm_sigma')
            self.modelpar.append('lognorm_ext')

            self.bounds['lognorm_radius'] = (np.log10(self.bounds['lognorm_radius'][0]),
                                             np.log10(self.bounds['lognorm_radius'][1]))

        elif 'powerlaw_max' in self.bounds and 'powerlaw_exp' in self.bounds and \
                'powerlaw_ext' in self.bounds:

            self.cross_sections, _, _ = dust_util.interp_powerlaw(
                inc_phot, inc_spec, self.spectrum)

            self.modelpar.append('powerlaw_max')
            self.modelpar.append('powerlaw_exp')
            self.modelpar.append('powerlaw_ext')

            self.bounds['powerlaw_max'] = (np.log10(self.bounds['powerlaw_max'][0]),
                                           np.log10(self.bounds['powerlaw_max'][1]))

        else:
            self.cross_sections = None

        if 'ism_ext' in self.bounds:
            self.modelpar.append('ism_ext')

        if 'ism_red' in self.bounds:
            self.modelpar.append('ism_red')

        self.fix_param = {}
        del_param = []

        for key, value in self.bounds.items():
            if value[0] == value[1]:
                self.fix_param[key] = value[0]
                del_param.append(key)

        if del_param:
            print(f'Fixing {len(del_param)} parameters:')

            for item in del_param:
                print(f'   - {item} = {self.fix_param[item]}')

                self.modelpar.remove(item)
                del self.bounds[item]

        print(f'Fitting {len(self.modelpar)} parameters:')

        for item in self.modelpar:
            print(f'   - {item}')

        print('Prior boundaries:')

        for key, value in self.bounds.items():
            print(f'   - {key} = {value}')

    @typechecked
    def run_mcmc(self,
                 tag: str,
                 guess: Optional[Dict[str, Union[Optional[float],
                                                 List[Optional[float]],
                                                 Tuple[Optional[float],
                                                       Optional[float]]]]],
                 nwalkers: int = 200,
                 nsteps: int = 1000,
                 prior: Optional[Dict[str, Tuple[float, float]]] = None) -> None:
        """
        Function to run the MCMC sampler of ``emcee``. The functionalities of ``run_mcmc`` are
        more limited than :func:`~species.analysis.fit_model.FitModel.run_multinest` and
        ``run_multinest`` provides more robust results when sampling multimodal posterior
        distributions and has the additional advantage of returning the model evidence. Therefore,
        it is recommended to use ``run_multinest`` instead of ``run_mcmc``.

        Parameters
        ----------
        tag : str
            Database tag where the samples will be stored.
        guess : dict, None
            Guess for each parameter to initialize the walkers. Random values between the
            ``bounds`` are used is set to ``None``.
        nwalkers : int
            Number of walkers.
        nsteps : int
            Number of steps per walker.
        prior : dict(str, tuple(float, float)), None
            Dictionary with Gaussian priors for one or multiple parameters. The prior can be set
            for any of the atmosphere or calibration parameters, e.g.
            ``prior={'teff': (1200., 100.)}``. Additionally, a prior can be set for the mass, e.g.
            ``prior={'mass': (13., 3.)}`` for an expected mass of 13 Mjup with an uncertainty of
            3 Mjup. The parameter is not used if set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        print('Running MCMC...')

        ndim = len(self.bounds)

        if self.model == 'planck':

            if 'teff' in self.bounds:
                sigma = {'teff': 5., 'radius': 0.01}

            else:
                sigma = {}

                for i, item in enumerate(guess['teff']):
                    sigma[f'teff_{i}'] = 5.
                    guess[f'teff_{i}'] = guess['teff'][i]

                for i, item in enumerate(guess['radius']):
                    sigma[f'radius_{i}'] = 0.01
                    guess[f'radius_{i}'] = guess['radius'][i]

                del guess['teff']
                del guess['radius']

        else:
            sigma = {'teff': 5.,
                     'logg': 0.01,
                     'feh': 0.01,
                     'fsed': 0.01,
                     'co': 0.01,
                     'radius': 0.01}

        for item in self.spectrum:
            if item in self.fit_corr:
                sigma[f'corr_len_{item}'] = 0.01  # (dex)
                guess[f'corr_len_{item}'] = None  # (um)

                sigma[f'corr_amp_{item}'] = 0.1
                guess[f'corr_amp_{item}'] = None

        for item in self.spectrum:
            if f'scaling_{item}' in self.bounds:
                sigma[f'scaling_{item}'] = 0.01
                guess[f'scaling_{item}'] = guess[item][0]

            if f'error_{item}' in self.bounds:
                sigma[f'error_{item}'] = 0.1
                guess[f'error_{item}'] = guess[item][1]

            if item in guess:
                del guess[item]

        initial = np.zeros((nwalkers, ndim))

        for i, item in enumerate(self.modelpar):
            if guess[item] is not None:
                initial[:, i] = guess[item] + np.random.normal(0, sigma[item], nwalkers)

            else:
                initial[:, i] = np.random.uniform(low=self.bounds[item][0],
                                                  high=self.bounds[item][1],
                                                  size=nwalkers)

        # Create a dictionary with the indices of the parameters

        param_index = {}
        for i, item in enumerate(self.modelpar):
            param_index[item] = i

        with Pool(processes=cpu_count()):

            ens_sampler = emcee.EnsembleSampler(nwalkers,
                                                ndim,
                                                lnprob,
                                                args=([self.bounds,
                                                       self.model,
                                                       param_index,
                                                       self.objphot,
                                                       self.distance,
                                                       prior,
                                                       self.spectrum,
                                                       self.modelphot,
                                                       self.modelspec,
                                                       self.n_planck,
                                                       self.fit_corr]))

            ens_sampler.run_mcmc(initial, nsteps, progress=True)

        spec_labels = []
        for item in self.spectrum:
            if f'scaling_{item}' in self.bounds:
                spec_labels.append(f'scaling_{item}')

        species_db = database.Database()

        species_db.add_samples(sampler='emcee',
                               samples=ens_sampler.chain,
                               ln_prob=ens_sampler.lnprobability,
                               mean_accept=np.mean(ens_sampler.acceptance_fraction),
                               spectrum=('model', self.model),
                               tag=tag,
                               modelpar=self.modelpar,
                               distance=self.distance[0],
                               spec_labels=spec_labels)

    @typechecked
    def run_multinest(self,
                      tag: str,
                      n_live_points: int = 1000,
                      output: str = 'multinest/',
                      prior: Optional[Dict[str, Tuple[float, float]]] = None) -> None:
        """
        Function to run the ``PyMultiNest`` wrapper of the ``MultiNest`` sampler. While
        ``PyMultiNest`` can be installed with ``pip`` from the PyPI repository, ``MultiNest``
        has to to be build manually. See the ``PyMultiNest`` documentation for details:
        http://johannesbuchner.github.io/PyMultiNest/install.html. Note that the library path
        of ``MultiNest`` should be set to the environmental variable ``LD_LIBRARY_PATH`` on a
        Linux machine and ``DYLD_LIBRARY_PATH`` on a Mac. Alternatively, the variable can be
        set before importing the ``species`` package, for example:

        .. code-block:: python

            >>> import os
            >>> os.environ['DYLD_LIBRARY_PATH'] = '/path/to/MultiNest/lib'
            >>> import species

        Parameters
        ----------
        tag : str
            Database tag where the samples will be stored.
        n_live_points : int
            Number of live points.
        output : str
            Path that is used for the output files from MultiNest.
        prior : dict(str, tuple(float, float)), None
            Dictionary with Gaussian priors for one or multiple parameters. The prior can be set
            for any of the atmosphere or calibration parameters, e.g.
            ``prior={'teff': (1200., 100.)}``. Additionally, a prior can be set for the mass, e.g.
            ``prior={'mass': (13., 3.)}`` for an expected mass of 13 Mjup with an uncertainty of
            3 Mjup. The parameter is not used if set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        print('Running nested sampling...')

        # Create the output folder if required

        if not os.path.exists(output):
            os.mkdir(output)

        # Create a dictionary with the cube indices of the parameters

        cube_index = {}
        for i, item in enumerate(self.modelpar):
            cube_index[item] = i

        @typechecked
        def lnprior_multinest(cube,
                              n_dim: int,
                              n_param: int) -> None:
            """
            Function to transform the unit cube into the parameter cube. It is not clear how to
            pass additional arguments to the function, therefore it is placed here and not merged
            with :func:`~species.analysis.fit_model.FitModel.run_mcmc`.

            Parameters
            ----------
            cube : pymultinest.run.LP_c_double
                Unit cube.
            n_dim : int
                Number of dimensions.
            n_param : int
                Number of parameters.

            Returns
            -------
            NoneType
                None
            """

            for item in cube_index:
                # Uniform priors for all parameters
                cube[cube_index[item]] = self.bounds[item][0] + \
                    (self.bounds[item][1]-self.bounds[item][0])*cube[cube_index[item]]

        @typechecked
        def lnlike_multinest(cube,
                             n_dim: int,
                             n_param: int) -> np.float64:
            """
            Function for calculating the log-likelihood for the sampled parameter cube.

            Parameters
            ----------
            cube : pymultinest.run.LP_c_double
                Unit cube.
            n_dim : int
                Number of dimensions. This parameter is mandatory but not used by the function.
            n_param : int
                Number of parameters. This parameter is mandatory but not used by the function.

            Returns
            -------
            float
                Log-likelihood.
            """

            # Initilize dictionaries for different parameter types

            spec_scaling = {}
            err_scaling = {}
            corr_len = {}
            corr_amp = {}
            dust_param = {}
            disk_param = {}
            param_dict = {}

            for item in self.bounds:
                # Add the parameters from the cube to their dictionaries

                if item[:8] == 'scaling_' and item[8:] in self.spectrum:
                    spec_scaling[item[8:]] = cube[cube_index[item]]

                elif item[:6] == 'error_' and item[6:] in self.spectrum:
                    err_scaling[item[6:]] = cube[cube_index[item]]

                elif item[:9] == 'corr_len_' and item[9:] in self.spectrum:
                    corr_len[item[9:]] = 10.**cube[cube_index[item]]  # (um)

                elif item[:9] == 'corr_amp_' and item[9:] in self.spectrum:
                    corr_amp[item[9:]] = cube[cube_index[item]]

                elif item[:8] == 'lognorm_':
                    dust_param[item] = cube[cube_index[item]]

                elif item[:9] == 'powerlaw_':
                    dust_param[item] = cube[cube_index[item]]

                elif item[:4] == 'ism_':
                    dust_param[item] = cube[cube_index[item]]

                elif item == 'disk_teff':
                    disk_param['teff'] = cube[cube_index[item]]

                elif item == 'disk_radius':
                    disk_param['radius'] = cube[cube_index[item]]

                else:
                    param_dict[item] = cube[cube_index[item]]

            for item in self.fix_param:
                # Add the fixed parameters to their dictionaries

                if item[:8] == 'scaling_' and item[8:] in self.spectrum:
                    spec_scaling[item[8:]] = self.fix_param[item]

                elif item[:6] == 'error_' and item[6:] in self.spectrum:
                    err_scaling[item[6:]] = self.fix_param[item]

                elif item[:9] == 'corr_len_' and item[9:] in self.spectrum:
                    corr_len[item[9:]] = self.fix_param[item]  # (um)

                elif item[:9] == 'corr_amp_' and item[9:] in self.spectrum:
                    corr_amp[item[9:]] = self.fix_param[item]

                elif item[:8] == 'lognorm_':
                    dust_param[item] = self.fix_param[item]

                elif item[:9] == 'powerlaw_':
                    dust_param[item] = self.fix_param[item]

                elif item[:4] == 'ism_':
                    dust_param[item] = self.fix_param[item]

                elif item == 'disk_teff':
                    disk_param['teff'] = self.fix_param[item]

                elif item == 'disk_radius':
                    disk_param['radius'] = self.fix_param[item]

                else:
                    param_dict[item] = self.fix_param[item]

            if self.model == 'planck' and self.n_planck > 1:
                for i in range(self.n_planck-1):
                    if param_dict[f'teff_{i+1}'] > param_dict[f'teff_{i}']:
                        return -np.inf

                    if param_dict[f'radius_{i}'] > param_dict[f'radius_{i+1}']:
                        return -np.inf

            if disk_param:
                if disk_param['teff'] > param_dict['teff']:
                    return -np.inf

                if disk_param['radius'] < param_dict['radius']:
                    return -np.inf

            if self.model == 'planck':
                param_dict['distance'] = self.distance[0]

            elif self.model != 'powerlaw':
                flux_scaling = (param_dict['radius']*constants.R_JUP)**2 / \
                               (self.distance[0]*constants.PARSEC)**2

                # The scaling is applied manually because of the interpolation
                del param_dict['radius']

            for item in self.spectrum:
                if item not in spec_scaling:
                    spec_scaling[item] = 1.

                if item not in err_scaling:
                    err_scaling[item] = None

            if self.param_interp is not None:
                # Sort the parameters in the correct order for spectrum_interp because
                # spectrum_interp creates a list in the order of the keys in param_dict
                param_tmp = param_dict.copy()

                param_dict = {}
                for item in self.param_interp:
                    param_dict[item] = param_tmp[item]

            ln_like = 0.

            if prior is not None:
                for key, value in prior.items():
                    if key == 'mass':
                        mass = read_util.get_mass(cube[cube_index['logg']],
                                                  cube[cube_index['radius']])

                        ln_like += -0.5 * (mass - value[0])**2 / value[1]**2

                    else:
                        ln_like += -0.5 * (cube[cube_index[key]] - value[0])**2 / value[1]**2

            if 'lognorm_ext' in dust_param:
                cross_tmp = self.cross_sections['Generic/Bessell.V'](
                    dust_param['lognorm_sigma'], 10.**dust_param['lognorm_radius'])[0]

                n_grains = dust_param['lognorm_ext'] / cross_tmp / 2.5 / np.log10(np.exp(1.))

            elif 'powerlaw_ext' in dust_param:
                cross_tmp = self.cross_sections['Generic/Bessell.V'](
                    dust_param['powerlaw_exp'], 10.**dust_param['powerlaw_max'])

                n_grains = dust_param['powerlaw_ext'] / cross_tmp / 2.5 / np.log10(np.exp(1.))

            for i, obj_item in enumerate(self.objphot):
                if self.modelphot[i] is not None:
                    phot_filter = self.modelphot[i].filter_name
                else:
                    phot_filter = None

                if self.model == 'planck':
                    readplanck = read_planck.ReadPlanck(filter_name=phot_filter)
                    phot_flux = readplanck.get_flux(param_dict, synphot=self.modelphot[i])[0]

                elif self.model == 'powerlaw':
                    powerl_box = read_util.powerlaw_spectrum(self.modelphot[i].wavel_range,
                                                             param_dict)

                    phot_flux = self.modelphot[i].spectrum_to_flux(
                        powerl_box.wavelength, powerl_box.flux)[0]

                else:
                    phot_flux = self.modelphot[i].spectrum_interp(list(param_dict.values()))[0][0]
                    phot_flux *= flux_scaling

                if disk_param:
                    phot_tmp = self.diskphot[i].spectrum_interp([disk_param['teff']])[0][0]

                    phot_flux += phot_tmp * (disk_param['radius']*constants.R_JUP)**2 / \
                        (self.distance[0]*constants.PARSEC)**2

                if 'lognorm_ext' in dust_param:
                    cross_tmp = self.cross_sections[phot_filter](
                        dust_param['lognorm_sigma'], 10.**dust_param['lognorm_radius'])[0]

                    phot_flux *= np.exp(-cross_tmp*n_grains)

                elif 'powerlaw_ext' in dust_param:
                    cross_tmp = self.cross_sections[phot_filter](
                        dust_param['powerlaw_exp'], 10.**dust_param['powerlaw_max'])[0]

                    phot_flux *= np.exp(-cross_tmp*n_grains)

                elif 'ism_ext' in dust_param:
                    read_filt = read_filter.ReadFilter(phot_filter)
                    filt_wavel = np.array([read_filt.mean_wavelength()])

                    ism_reddening = dust_param.get('ism_red', 3.1)

                    ext_filt = dust_util.ism_extinction(dust_param['ism_ext'],
                                                        ism_reddening,
                                                        filt_wavel)

                    phot_flux *= 10.**(-0.4*ext_filt[0])

                if obj_item.ndim == 1:
                    phot_var = obj_item[1]**2

                    if self.model == 'powerlaw' and f'{phot_filter}_error' in param_dict:
                        phot_var += param_dict[f'{phot_filter}_error']**2 * obj_item[0]**2

                    ln_like += -0.5 * (obj_item[0] - phot_flux)**2 / phot_var

                    # Only required when fitting an error inflation
                    ln_like += -0.5 * np.log(2.*np.pi*phot_var)

                else:
                    phot_var = obj_item[1, j]**2

                    if self.model == 'powerlaw' and f'{phot_filter}_error' in param_dict:
                        phot_var += param_dict[f'{phot_filter}_error']**2 * obj_item[0, j]**2

                    for j in range(obj_item.shape[1]):
                        ln_like += -0.5 * (obj_item[0, j] - phot_flux)**2 / phot_var

                        # Only required when fitting an error inflation
                        ln_like += -0.5 * np.log(2.*np.pi*phot_var)

            for i, item in enumerate(self.spectrum.keys()):
                # Calculate or interpolate the model spectrum

                if self.model == 'planck':
                    # Calculate a blackbody spectrum
                    readplanck = read_planck.ReadPlanck((0.9*self.spectrum[item][0][0, 0],
                                                         1.1*self.spectrum[item][0][-1, 0]))

                    model_box = readplanck.get_spectrum(param_dict, 1000., smooth=True)

                    # Resample the spectrum to the observed wavelengths
                    model_flux = spectres.spectres(self.spectrum[item][0][:, 0],
                                                   model_box.wavelength,
                                                   model_box.flux)

                else:
                    # Interpolate the model spectrum from the grid
                    model_flux = self.modelspec[i].spectrum_interp(list(param_dict.values()))[0, :]

                    # Scale the spectrum by (radius/distance)^2
                    model_flux *= flux_scaling

                # Scale the spectrum data
                data_flux = spec_scaling[item]*self.spectrum[item][0][:, 1]

                if err_scaling[item] is None:
                    # Variance without error inflation
                    data_var = self.spectrum[item][0][:, 2]**2

                else:
                    # Variance with error inflation (see Piette & Madhusudhan 2020)
                    data_var = self.spectrum[item][0][:, 2]**2 + (err_scaling[item]*model_flux)**2

                if self.spectrum[item][2] is not None:
                    # The inverted covariance matrix is available

                    if err_scaling[item] is None:
                        # Use the inverted covariance matrix directly
                        data_cov_inv = self.spectrum[item][2]

                    else:
                        # Ratio of the inflated and original uncertainties
                        sigma_ratio = np.sqrt(data_var) / self.spectrum[item][0][:, 2]
                        sigma_j, sigma_i = np.meshgrid(sigma_ratio, sigma_ratio)

                        # Calculate the inverted matrix of the inflated covariances
                        data_cov_inv = np.linalg.inv(self.spectrum[item][1]*sigma_i*sigma_j)

                if disk_param:
                    model_tmp = self.diskspec[i].spectrum_interp([disk_param['teff']])[0, :]

                    model_tmp *= (disk_param['radius']*constants.R_JUP)**2 / \
                        (self.distance[0]*constants.PARSEC)**2

                    model_flux += model_tmp

                if 'lognorm_ext' in dust_param:
                    for j, cross_item in enumerate(self.cross_sections[item]):
                        cross_tmp = cross_item(dust_param['lognorm_sigma'],
                                               10.**dust_param['lognorm_radius'])[0]

                        model_flux[j] *= np.exp(-cross_tmp*n_grains)

                elif 'powerlaw_ext' in dust_param:
                    for j, cross_item in enumerate(self.cross_sections[item]):
                        # For loop over all wavelengths of a spectrum
                        cross_tmp = cross_item(dust_param['powerlaw_exp'],
                                               10.**dust_param['powerlaw_max'])[0]

                        model_flux[j] *= np.exp(-cross_tmp*n_grains)

                elif 'ism_ext' in dust_param:
                    ism_reddening = dust_param.get('ism_red', 3.1)

                    ext_filt = dust_util.ism_extinction(dust_param['ism_ext'],
                                                        ism_reddening,
                                                        self.spectrum[item][0][:, 0])

                    model_flux *= 10.**(-0.4*ext_filt)

                if self.spectrum[item][2] is not None:
                    # Use the inverted covariance matrix
                    ln_like += -0.5 * np.dot(data_flux-model_flux,
                                             np.dot(data_cov_inv, data_flux-model_flux))

                    ln_like += -0.5 * np.nansum(np.log(2.*np.pi*data_var))

                else:
                    if item in self.fit_corr:
                        # Covariance model (Wang et al. 2020)
                        wavel = self.spectrum[item][0][:, 0]  # (um)
                        wavel_j, wavel_i = np.meshgrid(wavel, wavel)

                        error = np.sqrt(data_var)  # (W m-2 um-1)
                        error_j, error_i = np.meshgrid(error, error)

                        cov_matrix = corr_amp[item]**2 * error_i * error_j * \
                            np.exp(-(wavel_i-wavel_j)**2 / (2.*corr_len[item]**2)) + \
                            (1.-corr_amp[item]**2) * np.eye(wavel.shape[0])*error_i**2

                        dot_tmp = np.dot(data_flux-model_flux,
                                         np.dot(np.linalg.inv(cov_matrix), data_flux-model_flux))

                        ln_like += -0.5*dot_tmp - 0.5*np.nansum(np.log(2.*np.pi*data_var))

                    else:
                        # Calculate the chi-square without a covariance matrix
                        ln_like += np.nansum(-0.5 * (data_flux-model_flux)**2 / data_var -
                                             0.5 * np.log(2.*np.pi*data_var))

            return ln_like

        pymultinest.run(lnlike_multinest,
                        lnprior_multinest,
                        len(self.modelpar),
                        outputfiles_basename=output,
                        resume=False,
                        n_live_points=n_live_points)

        # Create the Analyzer object
        analyzer = pymultinest.analyse.Analyzer(len(self.modelpar), outputfiles_basename=output)

        # Get a dictionary with the ln(Z) and its errors, the individual modes and their parameters
        # quantiles of the parameter posteriors
        stats = analyzer.get_stats()

        # Nested sampling global log-evidence
        ln_z = stats['nested sampling global log-evidence']
        ln_z_error = stats['nested sampling global log-evidence error']
        print(f'Nested sampling global log-evidence: {ln_z:.2f} +/- {ln_z_error:.2f}')

        # Nested sampling global log-evidence
        ln_z = stats['nested importance sampling global log-evidence']
        ln_z_error = stats['nested importance sampling global log-evidence error']
        print(f'Nested importance sampling global log-evidence: {ln_z:.2f} +/- {ln_z_error:.2f}')

        # Get the best-fit (highest likelihood) point
        print('Sample with the highest likelihood:')
        best_params = analyzer.get_best_fit()

        max_lnlike = best_params['log_likelihood']
        print(f'   - Log-likelihood = {max_lnlike:.2f}')

        for i, item in enumerate(best_params['parameters']):
            print(f'   - {self.modelpar[i]} = {item:.2f}')

        # Get the posterior samples
        samples = analyzer.get_equal_weighted_posterior()

        spec_labels = []
        for item in self.spectrum:
            if f'scaling_{item}' in self.bounds:
                spec_labels.append(f'scaling_{item}')

        species_db = database.Database()

        ln_prob = samples[:, -1]
        samples = samples[:, :-1]

        # Adding the fixed parameters to the samples

        for key, value in self.fix_param.items():
            self.modelpar.append(key)

            app_param = np.full(samples.shape[0], value)
            app_param = app_param[..., np.newaxis]

            samples = np.append(samples, app_param, axis=1)

        species_db.add_samples(sampler='multinest',
                               samples=samples,
                               ln_prob=ln_prob,
                               mean_accept=None,
                               spectrum=('model', self.model),
                               tag=tag,
                               modelpar=self.modelpar,
                               distance=self.distance[0],
                               spec_labels=spec_labels)



"""
Module for downloading filter data from the website of the SVO Filter Profile Service.
"""

import os
import warnings
import urllib.request

from typing import Optional, Tuple
from astropy.io.votable import parse_single_table

import numpy as np

from typeguard import typechecked


@typechecked
def download_filter(filter_id: str) -> Tuple[Optional[np.ndarray],
                                             Optional[np.ndarray],
                                             Optional[str]]:
    """
    Function for downloading filter profile data from the SVO Filter Profile Service.

    Parameters
    ----------
    filter_id : str
        Filter name as listed on the website of the SVO Filter Profile Service
        (see http://svo2.cab.inta-csic.es/svo/theory/fps/).

    Returns
    -------
    np.ndarray
        Wavelength (um).
    np.ndarray
        Fractional transmission.
    str
        Detector type ('energy' or 'photon').
    """

    if filter_id == 'Magellan/VisAO.rp':
        url = 'https://xwcl.science/magao/visao/VisAO_rp_filter_curve.dat'
        urllib.request.urlretrieve(url, 'VisAO_rp_filter_curve.dat')

        wavelength, transmission, _, _ = np.loadtxt('VisAO_rp_filter_curve.dat', unpack=True)

        # Not sure if energy- or photon-counting detector
        det_type = 'energy'

        os.remove('VisAO_rp_filter_curve.dat')

    elif filter_id == 'Magellan/VisAO.ip':
        url = 'https://xwcl.science/magao/visao/VisAO_ip_filter_curve.dat'
        urllib.request.urlretrieve(url, 'VisAO_ip_filter_curve.dat')

        wavelength, transmission, _, _ = np.loadtxt('VisAO_ip_filter_curve.dat', unpack=True)

        # Not sure if energy- or photon-counting detector
        det_type = 'energy'

        os.remove('VisAO_ip_filter_curve.dat')

    elif filter_id == 'Magellan/VisAO.zp':
        url = 'https://xwcl.science/magao/visao/VisAO_zp_filter_curve.dat'
        urllib.request.urlretrieve(url, 'VisAO_zp_filter_curve.dat')

        wavelength, transmission, _, _ = np.loadtxt('VisAO_zp_filter_curve.dat', unpack=True)

        # Not sure if energy- or photon-counting detector
        det_type = 'energy'

        os.remove('VisAO_zp_filter_curve.dat')

    elif filter_id == 'LCO/VisAO.Ys' or filter_id == 'Magellan/VisAO.Ys':
        url = 'https://xwcl.science/magao/visao/VisAO_Ys_filter_curve.dat'
        urllib.request.urlretrieve(url, 'VisAO_Ys_filter_curve.dat')

        wavelength, transmission, _, _ = np.loadtxt('VisAO_Ys_filter_curve.dat', unpack=True)

        # Remove wavelengths with zero transmission
        wavelength = wavelength[:-7]
        transmission = transmission[:-7]

        # Not sure if energy- or photon-counting detector
        det_type = 'energy'

        os.remove('VisAO_Ys_filter_curve.dat')

    elif filter_id == 'ALMA/band6':
        url = 'https://home.strw.leidenuniv.nl/~stolker/species/alma_band6.dat'
        urllib.request.urlretrieve(url, 'alma_band6.dat')

        wavelength, transmission = np.loadtxt('alma_band6.dat', unpack=True)

        det_type = 'energy'

        os.remove('alma_band6.dat')

    elif filter_id == 'ALMA/band7':
        url = 'https://home.strw.leidenuniv.nl/~stolker/species/alma_band7.dat'
        urllib.request.urlretrieve(url, 'alma_band7.dat')

        wavelength, transmission = np.loadtxt('alma_band7.dat', unpack=True)

        det_type = 'energy'

        os.remove('alma_band7.dat')

    else:
        url = 'http://svo2.cab.inta-csic.es/svo/theory/fps/fps.php?ID='+filter_id
        urllib.request.urlretrieve(url, 'filter.xml')

        try:
            table = parse_single_table('filter.xml')

            wavelength = table.array['Wavelength']
            transmission = table.array['Transmission']

        except IndexError:
            wavelength = None
            transmission = None
            det_type = None

            warnings.warn(f'Filter \'{filter_id}\' is not available on the SVO Filter Profile '
                          f'Service.')

        except:
            os.remove('filter.xml')

            raise ValueError(f'The filter data of \'{filter_id}\' could not be downloaded. '
                             f'Perhaps the website of the SVO Filter Profile Service '
                             f'(http://svo2.cab.inta-csic.es/svo/theory/fps/) is not '
                             f'available?')

        if transmission is not None:
            try:
                det_type = table.get_field_by_id('DetectorType').value

                # For backward compatibility
                if not isinstance(det_type, str):
                    det_type = det_type.decode('utf-8')

                if int(det_type) == 1:
                    det_type = 'photon'

            except KeyError:
                # Energy-counting detector if the DetectorType key is not present
                det_type = 'energy'

            wavelength *= 1e-4  # (um)

        os.remove('filter.xml')

    if wavelength is not None:
        indices = []

        for i in range(transmission.size):
            if i == 0 and transmission[i] == 0. and transmission[i+1] == 0.:
                indices.append(i)

            elif i == transmission.size-1 and transmission[i-1] == 0. and transmission[i] == 0.:
                indices.append(i)

            elif transmission[i-1] == 0. and transmission[i] == 0. and transmission[i+1] == 0.:
                indices.append(i)

        wavelength = np.delete(wavelength, indices)
        transmission = np.delete(transmission, indices)

        if np.amin(transmission) < 0.:
            warnings.warn(f'The minimum transmission value of {filter_id} is smaller than zero '
                          f'({np.amin(transmission):.2e}). Wavelengths with negative transmission '
                          f'values will be removed.')

            indices = []

            for i, item in enumerate(transmission):
                if item > 0.:
                    indices.append(i)

            wavelength = wavelength[indices]
            transmission = transmission[indices]

    return wavelength, transmission, det_type



import os
import shutil

import pytest
import numpy as np

import species
from species.util import test_util


class TestObject:

    def setup_class(self):
        self.limit = 1e-8
        self.test_path = os.path.dirname(__file__) + '/'

    def teardown_class(self):
        os.remove('species_database.hdf5')
        os.remove('species_config.ini')
        shutil.rmtree('data/')

    def test_species_init(self):
        test_util.create_config('./')
        species.SpeciesInit()

    def test_read_object(self):
        database = species.Database()
        database.add_companion(name='beta Pic b')

        read_object = species.ReadObject('beta Pic b')
        assert read_object.object_name == 'beta Pic b'

        with pytest.raises(ValueError) as error:
            species.ReadObject('wrong name')

        assert str(error.value) == 'The object \'wrong name\' is not present in the database.'

    def test_get_photometry(self):
        read_object = species.ReadObject('beta Pic b')
        photometry = read_object.get_photometry('Paranal/NACO.Lp')

        assert isinstance(photometry, np.ndarray)

        assert photometry[0] == pytest.approx(11.3, rel=self.limit, abs=0.)
        assert photometry[1] == pytest.approx(0.06, rel=self.limit, abs=0.)
        assert photometry[2] == pytest.approx(1.5898815481080244e-15, rel=self.limit, abs=0.)
        assert photometry[3] == pytest.approx(8.790482731334405e-17, rel=self.limit, abs=0.)

    def test_get_distance(self):
        read_object = species.ReadObject('beta Pic b')
        distance = read_object.get_distance()

        assert distance == (19.75, 0.13)

    def test_get_absmag(self):
        read_object = species.ReadObject('beta Pic b')
        abs_mag = read_object.get_absmag('Paranal/NACO.Lp')

        assert abs_mag[0] == pytest.approx(9.822164500187604, rel=self.limit, abs=0.)
        assert abs_mag[1] == pytest.approx(0.06167898020063019, rel=self.limit, abs=0.)



"""
Module with a function for plotting spectra.
"""

import os
import math
import warnings
import itertools

from typing import Optional, Union, Tuple, List

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from typeguard import typechecked
from matplotlib.ticker import AutoMinorLocator, MultipleLocator

from species.core import box, constants
from species.read import read_filter
from species.util import plot_util


@typechecked
def plot_spectrum(boxes: list,
                  filters: Optional[List[str]] = None,
                  residuals: Optional[box.ResidualsBox] = None,
                  plot_kwargs: Optional[List[Optional[dict]]] = None,
                  xlim: Optional[Tuple[float, float]] = None,
                  ylim: Optional[Tuple[float, float]] = None,
                  ylim_res: Optional[Tuple[float, float]] = None,
                  scale: Optional[Tuple[str, str]] = None,
                  title: Optional[str] = None,
                  offset: Optional[Tuple[float, float]] = None,
                  legend: Optional[Union[str, dict, Tuple[float, float],
                                   List[Optional[Union[dict, str, Tuple[float, float]]]]]] = None,
                  figsize: Optional[Tuple[float, float]] = (10., 5.),
                  object_type: str = 'planet',
                  quantity: str = 'flux density',
                  output: str = 'spectrum.pdf'):
    """
    Parameters
    ----------
    boxes : list(species.core.box, )
        Boxes with data.
    filters : list(str, ), None
        Filter IDs for which the transmission profile is plotted. Not plotted if set to None.
    residuals : species.core.box.ResidualsBox, None
        Box with residuals of a fit. Not plotted if set to None.
    plot_kwargs : list(dict, ), None
        List with dictionaries of keyword arguments for each box. For example, if the ``boxes``
        are a ``ModelBox`` and ``ObjectBox``:

        .. code-block:: python

            plot_kwargs=[{'ls': '-', 'lw': 1., 'color': 'black'},
                         {'spectrum_1': {'marker': 'o', 'ms': 3., 'color': 'tab:brown', 'ls': 'none'},
                          'spectrum_2': {'marker': 'o', 'ms': 3., 'color': 'tab:blue', 'ls': 'none'},
                          'Paranal/SPHERE.IRDIS_D_H23_3': {'marker': 's', 'ms': 4., 'color': 'tab:cyan', 'ls': 'none'},
                          'Paranal/SPHERE.IRDIS_D_K12_1': [{'marker': 's', 'ms': 4., 'color': 'tab:orange', 'ls': 'none'},
                                                           {'marker': 's', 'ms': 4., 'color': 'tab:red', 'ls': 'none'}],
                          'Paranal/NACO.Lp': {'marker': 's', 'ms': 4., 'color': 'tab:green', 'ls': 'none'},
                          'Paranal/NACO.Mp': {'marker': 's', 'ms': 4., 'color': 'tab:green', 'ls': 'none'}}]

        For an ``ObjectBox``, the dictionary contains items for the different spectrum and filter
        names stored with :func:`~species.data.database.Database.add_object`. In case both
        and ``ObjectBox`` and a ``SynphotBox`` are provided, then the latter can be set to ``None``
        in order to use the same (but open) symbols as the data from the ``ObjectBox``. Note that
        if a filter name is duplicated in an ``ObjectBox`` (Paranal/SPHERE.IRDIS_D_K12_1 in the
        example) then a list with two dictionaries should be provided. Colors are automatically
        chosen if ``plot_kwargs`` is set to ``None``.
    xlim : tuple(float, float)
        Limits of the wavelength axis.
    ylim : tuple(float, float)
        Limits of the flux axis.
    ylim_res : tuple(float, float), None
        Limits of the residuals axis. Automatically chosen (based on the minimum and maximum
        residual value) if set to None.
    scale : tuple(str, str), None
        Scale of the x and y axes ('linear' or 'log'). The scale is set to ``('linear', 'linear')``
        if set to ``None``.
    title : str
        Title.
    offset : tuple(float, float)
        Offset for the label of the x- and y-axis.
    legend : str, tuple, dict, list(dict, dict), None
        Location of the legend (str or tuple(float, float)) or a dictionary with the ``**kwargs``
        of ``matplotlib.pyplot.legend``, for example ``{'loc': 'upper left', 'fontsize: 12.}``.
        Alternatively, a list with two values can be provided to separate the model and data
        handles in two legends. Each of these two elements can be set to ``None``. For example,
        ``[None, {'loc': 'upper left', 'fontsize: 12.}]``, if only the data points should be
        included in a legend.                  
    figsize : tuple(float, float)
        Figure size.
    object_type : str
        Object type ('planet' or 'star'). With 'planet', the radius and mass are expressed in
        Jupiter units. With 'star', the radius and mass are expressed in solar units.
    quantity: str
        The quantity of the y-axis ('flux density', 'flux', or 'magnitude').
    output : str
        Output filename.

    Returns
    -------
    NoneType
        None
    """

    mpl.rcParams['font.serif'] = ['Bitstream Vera Serif']
    mpl.rcParams['font.family'] = 'serif'

    plt.rc('axes', edgecolor='black', linewidth=2.2)
    plt.rcParams['axes.axisbelow'] = False

    if plot_kwargs is None:
        plot_kwargs = []

    elif plot_kwargs is not None and len(boxes) != len(plot_kwargs):
        raise ValueError(f'The number of \'boxes\' ({len(boxes)}) should be equal to the '
                         f'number of items in \'plot_kwargs\' ({len(plot_kwargs)}).')

    if residuals is not None and filters is not None:
        plt.figure(1, figsize=figsize)
        gridsp = mpl.gridspec.GridSpec(3, 1, height_ratios=[1, 3, 1])
        gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[1, 0])
        ax2 = plt.subplot(gridsp[0, 0])
        ax3 = plt.subplot(gridsp[2, 0])

    elif residuals is not None:
        plt.figure(1, figsize=figsize)
        gridsp = mpl.gridspec.GridSpec(2, 1, height_ratios=[4, 1])
        gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[0, 0])
        ax2 = None
        ax3 = plt.subplot(gridsp[1, 0])

    elif filters is not None:
        plt.figure(1, figsize=figsize)
        gridsp = mpl.gridspec.GridSpec(2, 1, height_ratios=[1, 4])
        gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[1, 0])
        ax2 = plt.subplot(gridsp[0, 0])
        ax3 = None

    else:
        plt.figure(1, figsize=figsize)
        gridsp = mpl.gridspec.GridSpec(1, 1)
        gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[0, 0])
        ax2 = None
        ax3 = None

    if residuals is not None:
        labelbottom = False
    else:
        labelbottom = True

    if scale is None:
        scale = ('linear', 'linear')

    ax1.set_xscale(scale[0])
    ax1.set_yscale(scale[1])

    if filters is not None:
        ax2.set_xscale(scale[0])

    if residuals is not None:
        ax3.set_xscale(scale[0])

    ax1.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                    direction='in', width=1, length=5, labelsize=12, top=True,
                    bottom=True, left=True, right=True, labelbottom=labelbottom)

    ax1.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                    direction='in', width=1, length=3, labelsize=12, top=True,
                    bottom=True, left=True, right=True, labelbottom=labelbottom)

    if filters is not None:
        ax2.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                        direction='in', width=1, length=5, labelsize=12, top=True,
                        bottom=True, left=True, right=True, labelbottom=False)

        ax2.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                        direction='in', width=1, length=3, labelsize=12, top=True,
                        bottom=True, left=True, right=True, labelbottom=False)

    if residuals is not None:
        ax3.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                        direction='in', width=1, length=5, labelsize=12, top=True,
                        bottom=True, left=True, right=True)

        ax3.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                        direction='in', width=1, length=3, labelsize=12, top=True,
                        bottom=True, left=True, right=True)

    if scale[0] == 'linear':
        ax1.xaxis.set_minor_locator(AutoMinorLocator(5))

    if scale[1] == 'linear':
        ax1.yaxis.set_minor_locator(AutoMinorLocator(5))

    # ax1.set_yticks([1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1e0])
    # ax3.set_yticks([-2., 0., 2.])

    if filters is not None and scale[0] == 'linear':
        ax2.xaxis.set_minor_locator(AutoMinorLocator(5))

    if residuals is not None and scale[0] == 'linear':
        ax3.xaxis.set_minor_locator(AutoMinorLocator(5))

    if residuals is not None and filters is not None:
        ax1.set_xlabel('')
        ax2.set_xlabel('')
        ax3.set_xlabel('Wavelength (µm)', fontsize=13)

    elif residuals is not None:
        ax1.set_xlabel('')
        ax3.set_xlabel('Wavelength (µm)', fontsize=11)

    elif filters is not None:
        ax1.set_xlabel('Wavelength (µm)', fontsize=13)
        ax2.set_xlabel('')

    else:
        ax1.set_xlabel('Wavelength (µm)', fontsize=13)

    if filters is not None:
        ax2.set_ylabel(r'T$_\lambda$', fontsize=13)

    if residuals is not None:
        if quantity == 'flux density':
            ax3.set_ylabel(r'$\Delta$$\mathregular{F}_\lambda$ ($\sigma$)', fontsize=11)

        elif quantity == 'flux':
            ax3.set_ylabel(r'$\Delta$$\mathregular{F}_\lambda$ ($\sigma$)', fontsize=11)

    if xlim is None:
        ax1.set_xlim(0.6, 6.)
    else:
        ax1.set_xlim(xlim[0], xlim[1])

    if quantity == 'magnitude':
        scaling = 1.
        ax1.set_ylabel('Flux contrast (mag)', fontsize=13)

        if ylim:
            ax1.set_ylim(ylim[0], ylim[1])

    else:
        if ylim:
            ax1.set_ylim(ylim[0], ylim[1])

            ylim = ax1.get_ylim()

            exponent = math.floor(math.log10(ylim[1]))
            scaling = 10.**exponent

            if quantity == 'flux density':
                ylabel = r'$\mathregular{F}_\lambda$ (10$^{'+str(exponent)+r'}$ W m$^{-2}$ µm$^{-1}$)'

            elif quantity == 'flux':
                ylabel = r'$\lambda$$\mathregular{F}_\lambda$ (10$^{'+str(exponent)+r'}$ W m$^{-2}$)'

            ax1.set_ylabel(ylabel, fontsize=11)
            ax1.set_ylim(ylim[0]/scaling, ylim[1]/scaling)

            if ylim[0] < 0.:
                ax1.axhline(0.0, ls='--', lw=0.7, color='gray', dashes=(2, 4), zorder=0.5)

        else:
            if quantity == 'flux density':
                ax1.set_ylabel(r'$\mathregular{F}_\lambda$ (W m$^{-2}$ µm$^{-1}$)', fontsize=11)

            elif quantity == 'flux':
                ax1.set_ylabel(r'$\lambda$$\mathregular{F}_\lambda$ (W m$^{-2}$)', fontsize=11)

            scaling = 1.

    xlim = ax1.get_xlim()

    if filters is not None:
        ax2.set_xlim(xlim[0], xlim[1])
        ax2.set_ylim(0., 1.)

    if residuals is not None:
        ax3.set_xlim(xlim[0], xlim[1])

    if offset is not None and residuals is not None and filters is not None:
        ax3.get_xaxis().set_label_coords(0.5, offset[0])

        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
        ax2.get_yaxis().set_label_coords(offset[1], 0.5)
        ax3.get_yaxis().set_label_coords(offset[1], 0.5)

    elif offset is not None and filters is not None:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])

        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
        ax2.get_yaxis().set_label_coords(offset[1], 0.5)

    elif offset is not None and residuals is not None:
        ax3.get_xaxis().set_label_coords(0.5, offset[0])

        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
        ax3.get_yaxis().set_label_coords(offset[1], 0.5)

    elif offset is not None:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.12)
        ax1.get_yaxis().set_label_coords(-0.1, 0.5)

    for j, boxitem in enumerate(boxes):
        flux_scaling = 1.

        if j < len(boxes):
            plot_kwargs.append(None)

        if isinstance(boxitem, (box.SpectrumBox, box.ModelBox)):
            wavelength = boxitem.wavelength
            flux = boxitem.flux

            if isinstance(wavelength[0], (np.float32, np.float64)):
                data = np.array(flux, dtype=np.float64)
                masked = np.ma.array(data, mask=np.isnan(data))

                if isinstance(boxitem, box.ModelBox):
                    param = boxitem.parameters

                    par_key, par_unit, par_label = plot_util.quantity_unit(
                        param=list(param.keys()), object_type=object_type)

                    label = ''
                    newline = False

                    for i, item in enumerate(par_key):
                        if item[:4] == 'teff':
                            value = f'{param[item]:.0f}'

                        elif item in ['logg', 'feh', 'fsed', 'lognorm_ext',
                                      'powerlaw_ext', 'ism_ext']:
                            value = f'{param[item]:.1f}'

                        elif item in ['co']:
                            value = f'{param[item]:.2f}'

                        elif item[:6] == 'radius':
                            if object_type == 'planet':
                                value = f'{param[item]:.1f}'

                                # if item == 'radius_1':
                                #     value = f'{param[item]:.0f}'
                                # else:
                                #     value = f'{param[item]:.1f}'

                            elif object_type == 'star':
                                value = f'{param[item]*constants.R_JUP/constants.R_SUN:.1f}'

                        elif item == 'mass':
                            if object_type == 'planet':
                                value = f'{param[item]:.0f}'

                            elif object_type == 'star':
                                value = f'{param[item]*constants.M_JUP/constants.M_SUN:.1f}'

                        elif item == 'luminosity':
                            value = f'{np.log10(param[item]):.2f}'

                        else:
                            continue

                        # if len(label) > 80 and newline == False:
                        #     label += '\n'
                        #     newline = True

                        if par_unit[i] is None:
                            label += f'{par_label[i]} = {value}'
                        else:
                            label += f'{par_label[i]} = {value} {par_unit[i]}'

                        if i < len(par_key)-1:
                            label += ', '

                else:
                    label = None

                if plot_kwargs[j]:
                    kwargs_copy = plot_kwargs[j].copy()

                    if 'label' in kwargs_copy:
                        if kwargs_copy['label'] is None:
                            label = None
                        else:
                            label = kwargs_copy['label']

                        del kwargs_copy['label']

                    if quantity == 'flux':
                        flux_scaling = wavelength

                    ax1.plot(wavelength, flux_scaling*masked/scaling, zorder=2, label=label, **kwargs_copy)

                else:
                    if quantity == 'flux':
                        flux_scaling = wavelength

                    ax1.plot(wavelength, flux_scaling*masked/scaling, lw=0.5, label=label, zorder=2)

            elif isinstance(wavelength[0], (np.ndarray)):
                for i, item in enumerate(wavelength):
                    data = np.array(flux[i], dtype=np.float64)
                    masked = np.ma.array(data, mask=np.isnan(data))

                    if isinstance(boxitem.name[i], bytes):
                        label = boxitem.name[i].decode('utf-8')
                    else:
                        label = boxitem.name[i]

                    if quantity == 'flux':
                        flux_scaling = item

                    ax1.plot(item, flux_scaling*masked/scaling, lw=0.5, label=label)

        elif isinstance(boxitem, list):
            for i, item in enumerate(boxitem):
                wavelength = item.wavelength
                flux = item.flux

                data = np.array(flux, dtype=np.float64)
                masked = np.ma.array(data, mask=np.isnan(data))

                if quantity == 'flux':
                    flux_scaling = wavelength

                if plot_kwargs[j]:
                    ax1.plot(wavelength, flux_scaling*masked/scaling, zorder=1, **plot_kwargs[j])
                else:
                    ax1.plot(wavelength, flux_scaling*masked/scaling, color='gray', lw=0.2, alpha=0.5, zorder=1)

        elif isinstance(boxitem, box.PhotometryBox):
            label_check = []

            for i, item in enumerate(boxitem.wavelength):
                transmission = read_filter.ReadFilter(boxitem.filter_name[i])
                fwhm = transmission.filter_fwhm()

                if quantity == 'flux':
                    flux_scaling = item

                if plot_kwargs[j]:
                    if 'label' in plot_kwargs[j] and plot_kwargs[j]['label'] not in label_check:
                        label_check.append(plot_kwargs[j]['label'])

                    elif 'label' in plot_kwargs[j] and plot_kwargs[j]['label'] in label_check:
                        del plot_kwargs[j]['label']

                    if boxitem.flux[i][1] is None:
                        ax1.errorbar(item, flux_scaling*boxitem.flux[i][0]/scaling, xerr=fwhm/2.,
                                     yerr=None, zorder=3, **plot_kwargs[j])

                    else:
                        ax1.errorbar(item, flux_scaling*boxitem.flux[i][0]/scaling, xerr=fwhm/2.,
                                     yerr=flux_scaling*boxitem.flux[i][1]/scaling, zorder=3, **plot_kwargs[j])

                else:
                    if boxitem.flux[i][1] is None:
                        ax1.errorbar(item, flux_scaling*boxitem.flux[i][0]/scaling, xerr=fwhm/2.,
                                     yerr=None, marker='s', ms=6, color='black', zorder=3)

                    else:
                        ax1.errorbar(item, flux_scaling*boxitem.flux[i][0]/scaling, xerr=fwhm/2.,
                                     yerr=flux_scaling*boxitem.flux[i][1]/scaling, marker='s', ms=6, color='black',
                                     zorder=3)

        elif isinstance(boxitem, box.ObjectBox):
            if boxitem.spectrum is not None:
                spec_list = []
                wavel_list = []

                for item in boxitem.spectrum:
                    spec_list.append(item)
                    wavel_list.append(boxitem.spectrum[item][0][0, 0])

                sort_index = np.argsort(wavel_list)
                spec_sort = []

                for i in range(sort_index.size):
                    spec_sort.append(spec_list[sort_index[i]])

                for key in spec_sort:
                    masked = np.ma.array(boxitem.spectrum[key][0],
                                         mask=np.isnan(boxitem.spectrum[key][0]))

                    if quantity == 'flux':
                        flux_scaling = masked[:, 0]

                    if not plot_kwargs[j] or key not in plot_kwargs[j]:
                        plot_obj = ax1.errorbar(masked[:, 0], flux_scaling*masked[:, 1]/scaling,
                                                yerr=flux_scaling*masked[:, 2]/scaling, ms=2, marker='s',
                                                zorder=2.5, ls='none')

                        if plot_kwargs[j] is None:
                            plot_kwargs[j] = {}

                        plot_kwargs[j][key] = {'marker': 's', 'ms': 2., 'ls': 'none',
                                               'color': plot_obj[0].get_color()}

                    else:
                        ax1.errorbar(masked[:, 0], flux_scaling*masked[:, 1]/scaling, yerr=flux_scaling*masked[:, 2]/scaling,
                                     zorder=2.5, **plot_kwargs[j][key])

            if boxitem.flux is not None:
                filter_list = []
                wavel_list = []

                for item in boxitem.flux:
                    read_filt = read_filter.ReadFilter(item)
                    filter_list.append(item)
                    wavel_list.append(read_filt.mean_wavelength())

                sort_index = np.argsort(wavel_list)
                filter_sort = []

                for i in range(sort_index.size):
                    filter_sort.append(filter_list[sort_index[i]])

                for item in filter_sort:
                    transmission = read_filter.ReadFilter(item)
                    wavelength = transmission.mean_wavelength()
                    fwhm = transmission.filter_fwhm()

                    if not plot_kwargs[j] or item not in plot_kwargs[j]:
                        if not plot_kwargs[j]:
                            plot_kwargs[j] = {}

                        if quantity == 'flux':
                            flux_scaling = wavelength

                        if isinstance(boxitem.flux[item][0], np.ndarray):
                            for i in range(boxitem.flux[item].shape[1]):

                                plot_obj = ax1.errorbar(wavelength, flux_scaling*boxitem.flux[item][0, i]/scaling, xerr=fwhm/2.,
                                             yerr=flux_scaling*boxitem.flux[item][1, i]/scaling, marker='s', ms=5, zorder=3)

                        else:
                            plot_obj = ax1.errorbar(wavelength, flux_scaling*boxitem.flux[item][0]/scaling, xerr=fwhm/2.,
                                         yerr=flux_scaling*boxitem.flux[item][1]/scaling, marker='s', ms=5, zorder=3)

                        plot_kwargs[j][item] = {'marker': 's', 'ms': 5., 'color': plot_obj[0].get_color()}

                    else:
                        if quantity == 'flux':
                            flux_scaling = wavelength

                        if isinstance(boxitem.flux[item][0], np.ndarray):
                            if not isinstance(plot_kwargs[j][item], list):
                                raise ValueError(f'A list with {boxitem.flux[item].shape[1]} '
                                                 f'dictionaries are required because the filter '
                                                 f'{item} has {boxitem.flux[item].shape[1]} '
                                                 f'values.')

                            for i in range(boxitem.flux[item].shape[1]):
                                ax1.errorbar(wavelength, flux_scaling*boxitem.flux[item][0, i]/scaling, xerr=fwhm/2.,
                                             yerr=flux_scaling*boxitem.flux[item][1, i]/scaling, zorder=3, **plot_kwargs[j][item][i])

                        else:
                            if boxitem.flux[item][1] == 0.:
                                ax1.errorbar(wavelength, flux_scaling*boxitem.flux[item][0]/scaling,
                                             xerr=fwhm/2., yerr=0.5*flux_scaling*boxitem.flux[item][0]/scaling,
                                             uplims=True, capsize=2., capthick=0., zorder=3, **plot_kwargs[j][item])

                            else:
                                ax1.errorbar(wavelength, flux_scaling*boxitem.flux[item][0]/scaling,
                                             xerr=fwhm/2., yerr=flux_scaling*boxitem.flux[item][1]/scaling,
                                             zorder=3, **plot_kwargs[j][item])

        elif isinstance(boxitem, box.SynphotBox):
            for i, find_item in enumerate(boxes):
                if isinstance(find_item, box.ObjectBox):
                    obj_index = i
                    break

            for item in boxitem.flux:
                transmission = read_filter.ReadFilter(item)
                wavelength = transmission.mean_wavelength()
                fwhm = transmission.filter_fwhm()

                if quantity == 'flux':
                    flux_scaling = wavelength

                if not plot_kwargs[obj_index] or item not in plot_kwargs[obj_index]:
                    ax1.errorbar(wavelength, flux_scaling*boxitem.flux[item]/scaling, xerr=fwhm/2., yerr=None,
                                 alpha=0.7, marker='s', ms=5, zorder=4, mfc='white')

                else:
                    if isinstance(plot_kwargs[obj_index][item], list):
                        # In case of multiple photometry values for the same filter, use the
                        # plot_kwargs of the first data point

                        kwargs_copy = plot_kwargs[obj_index][item][0].copy()

                        if 'label' in kwargs_copy:
                            del kwargs_copy['label']

                        ax1.errorbar(wavelength, flux_scaling*boxitem.flux[item]/scaling, xerr=fwhm/2., yerr=None,
                                     zorder=4, mfc='white', **kwargs_copy)

                    else:
                        kwargs_copy = plot_kwargs[obj_index][item].copy()

                        if 'label' in kwargs_copy:
                            del kwargs_copy['label']

                        if 'mfc' in kwargs_copy:
                            del kwargs_copy['mfc']

                        ax1.errorbar(wavelength, flux_scaling*boxitem.flux[item]/scaling, xerr=fwhm/2., yerr=None,
                                     zorder=4, mfc='white', **kwargs_copy)

    if filters is not None:
        for i, item in enumerate(filters):
            transmission = read_filter.ReadFilter(item)
            data = transmission.get_filter()

            ax2.plot(data[:, 0], data[:, 1], '-', lw=0.7, color='black', zorder=1)

    if residuals is not None:
        for i, find_item in enumerate(boxes):
            if isinstance(find_item, box.ObjectBox):
                obj_index = i
                break

        res_max = 0.

        if residuals.photometry is not None:
            for item in residuals.photometry:
                if not plot_kwargs[obj_index] or item not in plot_kwargs[obj_index]:
                    ax3.plot(residuals.photometry[item][0], residuals.photometry[item][1], marker='s',
                             ms=5, linestyle='none', zorder=2)

                else:
                    if residuals.photometry[item].ndim == 1:
                        ax3.errorbar(residuals.photometry[item][0], residuals.photometry[item][1],
                                     zorder=2, **plot_kwargs[obj_index][item])

                    elif residuals.photometry[item].ndim == 2:
                        for i in range(residuals.photometry[item].shape[1]):
                            if isinstance(plot_kwargs[obj_index][item], list):
                                ax3.errorbar(residuals.photometry[item][0, i],
                                             residuals.photometry[item][1, i], zorder=2,
                                             **plot_kwargs[obj_index][item][i])

                            else:
                                ax3.errorbar(residuals.photometry[item][0, i],
                                             residuals.photometry[item][1, i], zorder=2,
                                             **plot_kwargs[obj_index][item])

                res_max = np.nanmax(np.abs(residuals.photometry[item][1]))

        if residuals.spectrum is not None:
            for key, value in residuals.spectrum.items():
                if not plot_kwargs[obj_index] or key not in plot_kwargs[obj_index]:
                    ax3.errorbar(value[:, 0], value[:, 1], marker='o', ms=2, ls='none', zorder=1)

                else:
                    ax3.errorbar(value[:, 0], value[:, 1], zorder=1, **plot_kwargs[obj_index][key])

                max_tmp = np.nanmax(np.abs(value[:, 1]))

                if max_tmp > res_max:
                    res_max = max_tmp

        res_lim = math.ceil(1.1*res_max)

        if res_lim > 10.:
            res_lim = 5.

        ax3.axhline(0., ls='--', lw=0.7, color='gray', dashes=(2, 4), zorder=0.5)
        # ax3.axhline(-2.5, ls=':', lw=0.7, color='gray', dashes=(1, 4), zorder=0.5)
        # ax3.axhline(2.5, ls=':', lw=0.7, color='gray', dashes=(1, 4), zorder=0.5)

        if ylim_res is None:
            ax3.set_ylim(-res_lim, res_lim)

        else:
            ax3.set_ylim(ylim_res[0], ylim_res[1])

    if filters is not None:
        ax2.set_ylim(0., 1.1)

    print(f'Plotting spectrum: {output}...', end='', flush=True)

    if title is not None:
        if filters:
            ax2.set_title(title, y=1.02, fontsize=13)
        else:
            ax1.set_title(title, y=1.02, fontsize=13)

    handles, labels = ax1.get_legend_handles_labels()

    if handles and legend is not None:
        if isinstance(legend, list):
            model_handles = []
            data_handles = []

            model_labels = []
            data_labels = []

            for i, item in enumerate(handles):
                if isinstance(item, mpl.lines.Line2D):
                    model_handles.append(item)
                    model_labels.append(labels[i])

                elif isinstance(item, mpl.container.ErrorbarContainer):
                    data_handles.append(item)
                    data_labels.append(labels[i])

                else:
                    warnings.warn(f'The object type {item} is not implemented for the legend.')

            if legend[0] is not None:
                if isinstance(legend[0], (str, tuple)):
                    leg_1 = ax1.legend(model_handles, model_labels, loc=legend[0], fontsize=10., frameon=False)
                else:
                    leg_1 = ax1.legend(model_handles, model_labels, **legend[0])

            else:
                leg_1 = None

            if legend[1] is not None:
                if isinstance(legend[1], (str, tuple)):
                    leg_2 = ax1.legend(data_handles, data_labels, loc=legend[1], fontsize=8, frameon=False)
                else:
                    leg_2 = ax1.legend(data_handles, data_labels, **legend[1])

            if leg_1 is not None:
                ax1.add_artist(leg_1)

        elif isinstance(legend, (str, tuple)):
            ax1.legend(loc=legend, fontsize=8, frameon=False)

        else:
            ax1.legend(**legend)

    # filters = ['Paranal/SPHERE.ZIMPOL_N_Ha',
    #            'MUSE/Hbeta',
    #            'ALMA/855']
    #
    # filters = ['Paranal/SPHERE.IRDIS_B_Y',
    #            'MKO/NSFCam.J',
    #            'Paranal/SPHERE.IRDIS_D_H23_2',
    #            'Paranal/SPHERE.IRDIS_D_H23_3',
    #            'Paranal/SPHERE.IRDIS_D_K12_1',
    #            'Paranal/SPHERE.IRDIS_D_K12_2',
    #            'Paranal/NACO.Lp',
    #            'Paranal/NACO.NB405',
    #            'Paranal/NACO.Mp']
    #
    # for i, item in enumerate(filters):
    #     readfilter = read_filter.ReadFilter(item)
    #     filter_wavelength = readfilter.mean_wavelength()
    #     filter_width = readfilter.filter_fwhm()
    #
    #     # if i == 5:
    #     #     ax1.errorbar(filter_wavelength, 1.3e4, xerr=filter_width/2., color='dimgray', elinewidth=2.5, zorder=10)
    #     # else:
    #     #     ax1.errorbar(filter_wavelength, 6e3, xerr=filter_width/2., color='dimgray', elinewidth=2.5, zorder=10)
    #
    #     if i == 0:
    #         ax1.text(filter_wavelength, 1e-2, r'H$\alpha$', ha='center', va='center', fontsize=10, color='black')
    #     elif i == 1:
    #         ax1.text(filter_wavelength, 1e-2, r'H$\beta$', ha='center', va='center', fontsize=10, color='black')
    #     elif i == 2:
    #         ax1.text(filter_wavelength, 1e-2, 'ALMA\nband 7 rms', ha='center', va='center', fontsize=8, color='black')
    #
    #     if i == 0:
    #         ax1.text(filter_wavelength, 1.4, 'Y', ha='center', va='center', fontsize=10, color='black')
    #     elif i == 1:
    #         ax1.text(filter_wavelength, 1.4, 'J', ha='center', va='center', fontsize=10, color='black')
    #     elif i == 2:
    #         ax1.text(filter_wavelength-0.04, 1.4, 'H2', ha='center', va='center', fontsize=10, color='black')
    #     elif i == 3:
    #         ax1.text(filter_wavelength+0.04, 1.4, 'H3', ha='center', va='center', fontsize=10, color='black')
    #     elif i == 4:
    #         ax1.text(filter_wavelength, 1.4, 'K1', ha='center', va='center', fontsize=10, color='black')
    #     elif i == 5:
    #         ax1.text(filter_wavelength, 1.4, 'K2', ha='center', va='center', fontsize=10, color='black')
    #     elif i == 6:
    #         ax1.text(filter_wavelength, 1.4, 'L$\'$', ha='center', va='center', fontsize=10, color='black')
    #     elif i == 7:
    #         ax1.text(filter_wavelength, 1.4, 'NB4.05', ha='center', va='center', fontsize=10, color='black')
    #     elif i == 8:
    #         ax1.text(filter_wavelength, 1.4, 'M$\'}$', ha='center', va='center', fontsize=10, color='black')
    #
    # ax1.text(1.26, 0.58, 'VLT/SPHERE', ha='center', va='center', fontsize=8., color='slateblue', rotation=43.)
    # ax1.text(2.5, 1.28, 'VLT/SINFONI', ha='left', va='center', fontsize=8., color='darkgray')

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.clf()
    plt.close()

    print(' [DONE]')



"""
Module with functionalities for photometric and spectroscopic calibration. The fitting routine
can be used to fit photometric data with a calibration spectrum (e.g. extracted with
:func:`~species.read.read_model.ReadModel.get_model`) by simply fitting a scaling parameter.
"""

import math

from typing import Dict, List, Optional, Tuple, Union

from multiprocessing import cpu_count, Pool

import emcee
import numpy as np

from typeguard import typechecked

from species.analysis import photometry
from species.data import database
from species.read import read_calibration, read_object


@typechecked
def lnprob(param: np.ndarray,
           bounds: Dict[str, Tuple[float, float]],
           modelpar: List[str],
           objphot: List[np.ndarray],
           specphot: Union[List[float],
                           List[Tuple[photometry.SyntheticPhotometry,
                                      Tuple[np.float64, np.float64]]]]) -> float:
    """
    Internal function for calculating the posterior probability.

    Parameters
    ----------
    param : np.ndarray
        Value of the scaling parameter.
    bounds : dict
        Boundaries of the main scaling parameter.
    modelpar : list(str)
        Parameter names.
    objphot : list(tuple(float, float))
        Photometry of the object.
    specphot : list(float), photometry.SyntheticPhotometry
        Synthetic photometry of the calibration spectrum for the same filters as the photometry
        of the object.

    Returns
    -------
    float
        Log posterior probability.
    """

    ln_prob = 0.

    for i, item in enumerate(modelpar):

        if bounds[item][0] <= param[i] <= bounds[item][1]:
            ln_prob += 0.

        else:
            ln_prob += -np.inf
            break

    if not math.isinf(ln_prob):

        for i, obj_item in enumerate(objphot):
            if obj_item.ndim == 1:
                ln_prob += -0.5 * (obj_item[0] - param[0]*specphot[i])**2 / obj_item[1]**2

            else:

                for j in range(obj_item.shape[1]):
                    ln_prob += -0.5 * (obj_item[0, j] - param[0]*specphot[i])**2 / obj_item[1, j]**2

    return ln_prob


class FitSpectrum:
    """
    Class for fitting a calibration spectrum to photometric data.
    """

    @typechecked
    def __init__(self,
                 object_name: str,
                 filters: Optional[List[str]],
                 spectrum: str,
                 bounds: Dict[str, Tuple[float, float]]) -> None:
        """
        Parameters
        ----------
        object_name : str
            Object name in the database.
        filters : list(str)
            Filter names for which the photometry is selected. All available photometry of the
            object is selected if set to ``None``.
        spectrum : str
            Calibration spectrum as labelled in the database. The calibration spectrum can be
            stored in the database with :func:`~species.data.database.Database.add_calibration`.
        bounds : dict
            Boundaries of the scaling parameter, as ``{'scaling':(min, max)}``.

        Returns
        -------
        NoneType
            None
        """

        self.object = read_object.ReadObject(object_name)

        self.spectrum = spectrum
        self.bounds = bounds

        self.objphot = []
        self.specphot = []

        if filters is None:
            species_db = database.Database()

            objectbox = species_db.get_object(object_name, inc_phot=True, inc_spec=False)
            filters = objectbox.filters

        for item in filters:
            readcalib = read_calibration.ReadCalibration(self.spectrum, item)
            calibspec = readcalib.get_spectrum()

            synphot = photometry.SyntheticPhotometry(item)
            spec_phot = synphot.spectrum_to_flux(calibspec.wavelength, calibspec.flux)
            self.specphot.append(spec_phot[0])

            obj_phot = self.object.get_photometry(item)
            self.objphot.append(np.array([obj_phot[2], obj_phot[3]]))

        self.modelpar = ['scaling']

    @typechecked
    def run_mcmc(self,
                 nwalkers: int,
                 nsteps: int,
                 guess: Union[Dict[str, float], Dict[str, None]],
                 tag: str) -> None:
        """
        Function to run the MCMC sampler.

        Parameters
        ----------
        nwalkers : int
            Number of walkers.
        nsteps : int
            Number of steps per walker.
        guess : dict(str, float), dict(str, None)
            Guess of the scaling parameter.
        tag : str
            Database tag where the MCMC samples will be stored.

        Returns
        -------
        NoneType
            None
        """

        print('Running MCMC...')

        ndim = 1

        initial = np.zeros((nwalkers, ndim))

        for i, item in enumerate(self.modelpar):
            if guess[item] is not None:
                width = min(abs(guess[item] - self.bounds[item][0]),
                            abs(guess[item] - self.bounds[item][1]))

                initial[:, i] = guess[item] + np.random.normal(0, 0.1*width, nwalkers)

            else:
                initial[:, i] = np.random.uniform(low=self.bounds[item][0],
                                                  high=self.bounds[item][1],
                                                  size=nwalkers)

        with Pool(processes=cpu_count()):
            ens_sampler = emcee.EnsembleSampler(nwalkers,
                                                ndim,
                                                lnprob,
                                                args=([self.bounds,
                                                       self.modelpar,
                                                       self.objphot,
                                                       self.specphot]))

            ens_sampler.run_mcmc(initial, nsteps, progress=True)

        species_db = database.Database()

        species_db.add_samples(sampler='emcee',
                               samples=ens_sampler.chain,
                               ln_prob=ens_sampler.lnprobability,
                               mean_accept=np.mean(ens_sampler.acceptance_fraction),
                               spectrum=('calibration', self.spectrum),
                               tag=tag,
                               modelpar=self.modelpar,
                               distance=None,
                               spec_labels=None)



"""
Module for BT-NextGen atmospheric model spectra.
"""

import os
import tarfile
import urllib.request

from typing import Optional, Tuple

import h5py
import spectres
import numpy as np

from typeguard import typechecked

from species.util import data_util, read_util


@typechecked
def add_btnextgen(input_path: str,
                  database: h5py._hl.files.File,
                  wavel_range: Optional[Tuple[float, float]] = None,
                  teff_range: Optional[Tuple[float, float]] = None,
                  spec_res: float = None) -> None:
    """
    Function for adding the BT-NextGen atmospheric models to the database. The original spectra
    have been resampled to a spectral resolution of R = 2000 from 0.1 to 50 um.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    wavel_range : tuple(float, float), None
        Wavelength range (um). The full wavelength range (0.1-50 um) is stored if set to ``None``.
        Only used in combination with ``spec_res``.
    teff_range : tuple(float, float), None
        Effective temperature range (K). All available temperatures are stored if set to ``None``.
    spec_res : float, None
        Spectral resolution. The data is stored with the spectral resolution of the input spectra
        (R = 2000) if set to ``None``. Only used in combination with ``wavel_range``.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    input_file = 'bt-nextgen.tgz'
    url = 'https://people.phys.ethz.ch/~ipa/tstolker/bt-nextgen.tgz'

    data_folder = os.path.join(input_path, 'bt-nextgen/')
    data_file = os.path.join(input_path, input_file)

    if not os.path.exists(data_folder):
        os.makedirs(data_folder)

    if not os.path.isfile(data_file):
        print('Downloading BT-NextGen model spectra (368 MB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Unpacking BT-NextGen model spectra (368 MB)...', end='', flush=True)
    tar = tarfile.open(data_file)
    tar.extractall(data_folder)
    tar.close()
    print(' [DONE]')

    teff = []
    logg = []
    feh = []
    flux = []

    if wavel_range is not None and spec_res is not None:
        wavelength = read_util.create_wavelengths(wavel_range, spec_res)
    else:
        wavelength = None

    for _, _, files in os.walk(data_folder):
        for filename in files:
            if filename[:11] == 'bt-nextgen_':
                file_split = filename.split('_')

                teff_val = float(file_split[2])
                logg_val = float(file_split[4])
                feh_val = float(file_split[6])

                if teff_range is not None:
                    if teff_val < teff_range[0] or teff_val > teff_range[1]:
                        continue

                print_message = f'Adding BT-NextGen model spectra... {filename}'
                print(f'\r{print_message:<82}', end='')

                data_wavel, data_flux = np.loadtxt(os.path.join(data_folder, filename), unpack=True)

                teff.append(teff_val)
                logg.append(logg_val)
                feh.append(feh_val)

                if wavel_range is None or spec_res is None:
                    if wavelength is None:
                        wavelength = np.copy(data_wavel)  # (um)

                    if np.all(np.diff(wavelength) < 0):
                        raise ValueError('The wavelengths are not all sorted by increasing value.')

                    flux.append(data_flux)  # (W m-2 um-1)

                else:
                    flux_resample = spectres.spectres(wavelength,
                                                      data_wavel,
                                                      data_flux,
                                                      spec_errs=None,
                                                      fill=np.nan,
                                                      verbose=False)

                    if np.isnan(np.sum(flux_resample)):
                        raise ValueError(f'Resampling is only possible if the new wavelength '
                                         f'range ({wavelength[0]} - {wavelength[-1]} um) falls '
                                         f'sufficiently far within the wavelength range '
                                         f'({data_wavel[0]} - {data_wavel[-1]} um) of the input '
                                         f'spectra.')

                    flux.append(flux_resample)  # (W m-2 um-1)

    print_message = 'Adding BT-NextGen model spectra... [DONE]'
    print(f'\r{print_message:<82}')

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      None,
                                      None,
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('bt-nextgen',
                         ['teff', 'logg', 'feh'],
                         database,
                         data_sorted)
